"""
moptipy is the *Metaheuristic Optimization in Python* Package.

This is the basic package of our moptipy project.
It provides several sub-packages:

* :mod:`moptipy.algorithms` hosts implementations of several metaheuristic
  optimization algorithms
* :mod:`moptipy.api` defines the API for implementing optimization algorithms
  and -problems
* :mod:`moptipy.evaluation` offers evaluation utilities for the results of
  experiments
* :mod:`moptipy.mo` provides utilities for multi-objective optimization
* :mod:`moptipy.operators` implements search operators for various search
  spaces
* :mod:`moptipy.spaces` includes several pre-defined search spaces
* :mod:`moptipy.tests` provides several unit tests for moptipy components
* :mod:`moptipy.utils` has utility classes for, e.g., logging and naming
"""
from typing import Final

import moptipy.version

__version__: Final[str] = moptipy.version.__version__



"""Routines for handling strings."""

from re import compile as _compile
from re import sub
from typing import Final, Iterable, Pattern

from pycommons.strings.string_conv import float_to_str, num_to_str, str_to_num
from pycommons.strings.tools import replace_str
from pycommons.types import type_error


def num_to_str_for_name(x: int | float) -> str:
    """
    Convert a float to a string for use in a component name.

    This function can be inverted by applying :func:`name_str_to_num`.

    :param x: the float
    :returns: the string

    >>> num_to_str_for_name(1.3)
    '1d3'
    >>> num_to_str_for_name(1.0)
    '1'
    >>> num_to_str_for_name(-7)
    'm7'
    >>> num_to_str_for_name(-6.32)
    'm6d32'
    >>> num_to_str_for_name(-1e-5)
    'm1em5'
    """
    return num_to_str(x).replace(".", DECIMAL_DOT_REPLACEMENT) \
        .replace("-", MINUS_REPLACEMENT)


def name_str_to_num(s: str) -> int | float:
    """
    Convert a string from a name to a number.

    This function is the inverse of :func:`num_to_str_for_name`.

    :param s: the string from the name
    :returns: an integer or float, depending on the number represented by
        `s`

    >>> name_str_to_num(num_to_str_for_name(1.1))
    1.1
    >>> name_str_to_num(num_to_str_for_name(1))
    1
    >>> name_str_to_num(num_to_str_for_name(-5e3))
    -5000
    >>> name_str_to_num(num_to_str_for_name(-6e-3))
    -0.006
    >>> name_str_to_num(num_to_str_for_name(100.0))
    100
    >>> name_str_to_num(num_to_str_for_name(-1e-4))
    -0.0001
    """
    return str_to_num(s.replace(MINUS_REPLACEMENT, "-")
                      .replace(DECIMAL_DOT_REPLACEMENT, "."))


#: the internal table for converting normal characters to unicode superscripts
__SUPERSCRIPT: Final = str.maketrans({
    # numbers from 0 to 9
    0x30: 0x2070, 0x31: 0x00b9, 0x32: 0x00b2, 0x33: 0x00b3, 0x34: 0x2074,
    0x35: 0x2075, 0x36: 0x2076, 0x37: 0x2077, 0x38: 0x2078, 0x39: 0x2079,
    # +/-/=/(/)
    0x2b: 0x207A, 0x2d: 0x207b, 0x3d: 0x207c, 0x28: 0x207d, 0x29: 0x207e,
    # lower case letters
    0x61: 0x1d43, 0x62: 0x1d47, 0x63: 0x1d9c, 0x64: 0x1d48, 0x65: 0x1d49,
    0x66: 0x1da0, 0x67: 0x1d4d, 0x6b: 0x1d4f, 0x6c: 0x1da9, 0x6d: 0x1d50,
    0x6e: 0x207f, 0x6f: 0x1d52, 0x70: 0x1d56, 0x74: 0x1d57, 0x75: 0x1d58,
    0x76: 0x1d5b, 0x7a: 0x1dbb,
})


def superscript(s: str) -> str:
    """
    Transform a string into Unicode-based superscript.

    :param s: the string
    :returns: the string in superscript

    >>> superscript("a0=4(e)")
    '\u1d43\u2070\u207c\u2074\u207d\u1d49\u207e'
    """
    if not isinstance(s, str):
        raise type_error(s, "s", str)
    return s.translate(__SUPERSCRIPT)


def beautify_float_str(s: str | float) -> str:
    """
    Beautify the string representation of a float.

    This function beautifies the string representation of a float by using
    unicode superscripts for exponents.

    :param s: either a `float` or the string representation of a `float`
    :return: the beautified string representation

    >>> beautify_float_str('0.0')
    '0.0'
    >>> beautify_float_str('1e12')
    '1\u00d710\u00b9\u00b2'
    >>> beautify_float_str('1e-3')
    '1\u00d710\u207b\u00b3'
    >>> beautify_float_str('inf')
    '\u221e'
    >>> beautify_float_str('-inf')
    '-\u221e'
    >>> beautify_float_str('nan')
    '\u2205'
    """
    if isinstance(s, float):
        s = float_to_str(s)
    if not isinstance(s, str):
        raise type_error(s, "s", str)
    s = s.strip().lower()
    if s in ("+inf", "inf"):
        return "\u221e"
    if s == "-inf":
        return "-\u221e"
    if s == "nan":
        return "\u2205"
    eidx: int = s.find("e")
    if eidx < 0:
        return s
    return f"{s[:eidx]}\u00d710{s[eidx + 1:].translate(__SUPERSCRIPT)}"


def __replace_double(replace: str, src: str) -> str:
    """
    Replace any double-occurrence of a string with a single occurrence.

    :param replace: the string to replace
    :param src: the source string
    :returns: the updated string
    """
    return replace_str(replace + replace, replace, src)


#: the separator of different filename parts
PART_SEPARATOR: Final[str] = "_"
#: the replacement for "." in a file name
DECIMAL_DOT_REPLACEMENT: Final[str] = "d"
#: the replacement for "-" in a file name
MINUS_REPLACEMENT: Final[str] = "m"
#: the replacement for "+" in a file name
PLUS_REPLACEMENT: Final[str] = "p"

#: a pattern used during name sanitization
__PATTERN_SPACE_BEFORE_MINUS: Final[Pattern] = _compile(r"[^\w\s-]")
#: the multiple-whitespace pattern
__PATTERN_MULTIPLE_WHITESPACE: Final[Pattern] = _compile(r"\s+")


def sanitize_name(name: str) -> str:
    """
    Sanitize a name in such a way that it can be used as path component.

    >>> sanitize_name(" hello world ")
    'hello_world'
    >>> sanitize_name(" 56.6-455 ")
    '56d6m455'
    >>> sanitize_name(" _ i _ am _ funny   --6 _ ")
    'i_am_funny_m6'

    :param name: the name that should be sanitized
    :return: the sanitized name
    :raises ValueError: if the name is invalid or empty
    :raises TypeError: if the name is `None` or not a string
    """
    if not isinstance(name, str):
        raise type_error(name, "name", str)
    orig_name = name
    name = name.strip()
    name = __replace_double("-", name).replace("+", PLUS_REPLACEMENT)
    name = __replace_double("+", name).replace("-", MINUS_REPLACEMENT)
    name = __replace_double("_", name)
    name = __replace_double(".", name).replace(".", DECIMAL_DOT_REPLACEMENT)

    name = sub(__PATTERN_SPACE_BEFORE_MINUS, "", name)
    name = sub(__PATTERN_MULTIPLE_WHITESPACE, PART_SEPARATOR, name)
    name = __replace_double("_", name)

    if name.startswith("_"):
        name = name[1:]

    if name.endswith("_"):
        name = name[:len(name) - 1]

    if len(name) <= 0:
        raise ValueError(
            f"Sanitized name must not become empty, but {orig_name!r} does.")

    return name


def sanitize_names(names: Iterable[str]) -> str:
    """
    Sanitize a set of names.

    >>> sanitize_names(["", " sdf ", "", "5-3"])
    'sdf_5m3'
    >>> sanitize_names([" a ", " b", " c", "", "6", ""])
    'a_b_c_6'

    :param names: the list of names.
    :return: the sanitized name
    """
    return PART_SEPARATOR.join([
        sanitize_name(name) for name in names if len(name) > 0])



"""
The experiment execution API.

Via the function :func:`run_experiment`, you can execute a complex experiment
where multiple optimization algorithms are applied to multiple problem
instances, where log files with the results and progress information about the
runs are collected, and where multiprocessing is used to parallelize the
experiment execution.
Experiments are replicable, as random seeds are automatically generated based
on problem instance names in a replicable fashion.

The log files are structured according to the documentation in
https://thomasweise.github.io/moptipy/#file-names-and-folder-structure
and their contents follow the specification given in
https://thomasweise.github.io/moptipy/#log-file-sections.
"""
import copy
import gc
import multiprocessing as mp
import os.path
import platform
from contextlib import AbstractContextManager, nullcontext
from enum import IntEnum
from math import ceil
from typing import Any, Callable, Final, Iterable, Sequence, cast

import psutil  # type: ignore
from numpy.random import Generator, default_rng
from pycommons.ds.cache import str_is_new
from pycommons.io.console import logger
from pycommons.io.path import Path
from pycommons.types import check_int_range, type_error

from moptipy.api.execution import Execution
from moptipy.api.logging import FILE_SUFFIX
from moptipy.api.process import Process
from moptipy.utils.nputils import rand_seeds_from_str
from moptipy.utils.strings import sanitize_name, sanitize_names
from moptipy.utils.sys_info import get_sys_info, update_sys_info_cpu_affinity


def __run_experiment(base_dir: Path,
                     experiments: list[list[Callable]],
                     n_runs: list[int],
                     perform_warmup: bool,
                     warmup_fes: int,
                     perform_pre_warmup: bool,
                     pre_warmup_fes: int,
                     file_lock: AbstractContextManager,
                     stdio_lock: AbstractContextManager,
                     cache: Callable[[str], bool],
                     thread_id: str,
                     pre_warmup_barrier,
                     on_completion: Callable[[
                         Any, Path, Process], None]) -> None:
    """
    Execute a single thread of experiments.

    :param base_dir: the base directory
    :param experiments: the stream of experiment setups
    :param perform_warmup: should we perform a warm-up per instance?
    :param warmup_fes: the number of the FEs for the warm-up runs
    :param perform_pre_warmup: should we do one warmup run for each
        instance before we begin with the actual experiments?
    :param pre_warmup_fes: the FEs for the pre-warmup runs
    :param file_lock: the lock for file operations
    :param stdio_lock: the lock for log output
    :param cache: the cache
    :param thread_id: the thread id
    :param pre_warmup_barrier: a barrier to wait at after the pre-warmup
    :param on_completion: a function to be called for every completed run,
        receiving the instance, the path to the log file (before it is
        created) and the :class:`~moptipy.api.process.Process` of the run
        as parameters
    """
    random: Final[Generator] = default_rng()

    for warmup in ([True, False] if perform_pre_warmup else [False]):
        wss: str
        if warmup:
            wss = "pre-warmup"
        else:
            wss = "warmup"
            if perform_pre_warmup:
                gc.collect()  # do full garbage collection after pre-warmups
                gc.collect()  # one more, to be double-safe
                gc.freeze()  # whatever survived now, keep it permanently
                if pre_warmup_barrier:
                    logger(
                        "reached pre-warmup barrier.", thread_id, stdio_lock)
                    pre_warmup_barrier.wait()  # wait for all threads

        for runs in ([1] if warmup else n_runs):  # for each number of runs
            random.shuffle(cast(Sequence, experiments))  # shuffle experiments

            for setup in experiments:  # for each setup
                instance = setup[0]()  # load instance
                if instance is None:
                    raise TypeError("None is not an instance.")
                inst_name = sanitize_name(str(instance))

                exp = setup[1](instance)  # setup algorithm for instance
                if not isinstance(exp, Execution):
                    raise type_error(exp, "result of setup callable",
                                     Execution)
                # noinspection PyProtectedMember
                algo_name = sanitize_name(str(exp._algorithm))

                cd = Path(os.path.join(base_dir, algo_name, inst_name))
                cd.ensure_dir_exists()

                # generate sequence of seeds
                seeds: list[int] = [0] if warmup else \
                    rand_seeds_from_str(string=inst_name, n_seeds=runs)
                random.shuffle(seeds)
                needs_warmup = warmup or perform_warmup
                for seed in seeds:  # for every run

                    filename = sanitize_names(
                        [algo_name, inst_name, hex(seed)])
                    if warmup:
                        log_file = filename
                    else:
                        log_file = Path(
                            os.path.join(cd, filename + FILE_SUFFIX))

                        skip = True
                        with file_lock:
                            if cache(log_file):
                                skip = log_file.ensure_file_exists()
                        if skip:
                            continue  # run already done

                    exp.set_rand_seed(seed)

                    if needs_warmup:  # perform warmup run
                        needs_warmup = False
                        cpy: Execution = copy.copy(exp)
                        cpy.set_max_fes(
                            pre_warmup_fes if warmup else warmup_fes, True)
                        cpy.set_max_time_millis(3600000, True)
                        cpy.set_log_file(None)
                        cpy.set_log_improvements(False)
                        cpy.set_log_all_fes(False)
                        logger(
                            f"{wss} for {filename!r}.", thread_id, stdio_lock)
                        with cpy.execute():
                            pass
                        del cpy

                    if warmup:
                        continue

                    exp.set_log_file(log_file)
                    logger(filename, thread_id, stdio_lock)
                    with exp.execute() as process:  # run the experiment
                        on_completion(instance, cast(Path, log_file), process)


#: the number of logical CPU cores
_CPU_LOGICAL_CORES: Final[int] = psutil.cpu_count(logical=True)
#: the number of phyiscal CPU cores
_CPU_PHYSICAL_CORES: Final[int] = psutil.cpu_count(logical=False)
#: the logical cores per physical core
_CPU_LOGICAL_PER_PHYSICAL: Final[int] = \
    max(1, int(ceil(_CPU_LOGICAL_CORES / _CPU_PHYSICAL_CORES)))


class Parallelism(IntEnum):
    """
    An enumeration of parallel thread counts.

    The `value` of each element of this enumeration equal the number of
    threads that would be used. Thus, the values are different on
    different systems. Currently, only Linux is supported for parallelism.
    For all other systems, you need to manually start the program as often
    as you want it to run in parallel. On other systems, all values of this
    enumeration are `1`.

    >>> Parallelism.SINGLE_THREAD.value
    1
    """

    #: use only a single thread
    SINGLE_THREAD = 1
    #: Use as many threads as accurate time measurement permits. This equals
    #: to using one logical core on each physical core and leaving one
    #: physical core unoccupied (but always using at least one thread,
    #: obviously). If you have four physical cores with two logical cores
    #: each, this would mean using three threads on Linux.
    #: On Windows, parallelism is not supported yet, so this would equal using
    #: only one core on Windows.
    ACCURATE_TIME_MEASUREMENTS = max(1, _CPU_PHYSICAL_CORES - 1) \
        if "Linux" in platform.system() else 1
    #: Use all but one logical core. This *will* mess up time measurements but
    #: should produce the maximum performance while not impeding the system of
    #: doing stuff like garbage collection or other bookkeeping and overhead
    #: tasks. This is the most reasonable option if you want to execute one
    #: experiment as quickly as possible. If you have four physical cores with
    #: two logical cores each, this would mean using seven threads on Linux.
    #: On Windows, parallelism is not supported yet, so this would equal using
    #: only one core on Windows.
    PERFORMANCE = max(1, _CPU_LOGICAL_CORES - 1) \
        if "Linux" in platform.system() else 1
    #: Use every single logical core available, which may mess up your system.
    #: We run the experiment as quickly as possible, but the system may not be
    #: usable while the experiment is running. Background tasks like garbage
    #: collection, moving the mouse course, accepting user input, or network
    #: communication may come to a halt. Seriously, why would you use this?
    #: If you have four physical cores with  two logical cores each, this would
    #: mean using eight threads on Linux.
    #: On Windows, parallelism is not supported yet, so this would equal using
    #: only one core on Windows.
    RECKLESS = max(1, _CPU_LOGICAL_CORES) \
        if "Linux" in platform.system() else 1


def __waiting_run_experiment(
        base_dir: Path, experiments: list[list[Callable]],
        n_runs: list[int], perform_warmup: bool, warmup_fes: int,
        perform_pre_warmup: bool, pre_warmup_fes: int,
        file_lock: AbstractContextManager,
        stdio_lock: AbstractContextManager, cache: Callable, thread_id: str,
        event, pre_warmup_barrier,
        on_completion: Callable[[Any, Path, Process], None]) -> None:
    """Wait until event is set, then run experiment."""
    logger("waiting for start signal", thread_id, stdio_lock)
    if not event.wait():
        raise ValueError("Wait terminated unexpectedly.")
    logger("got start signal, beginning experiment", thread_id, stdio_lock)
    update_sys_info_cpu_affinity()
    gc.collect()
    __run_experiment(base_dir, experiments, n_runs,
                     perform_warmup, warmup_fes, perform_pre_warmup,
                     pre_warmup_fes, file_lock, stdio_lock, cache,
                     thread_id, pre_warmup_barrier, on_completion)


def __no_complete(_: Any, __: Path, ___: Process) -> None:
    """Do nothing."""


def run_experiment(
        base_dir: str, instances: Iterable[Callable[[], Any]],
        setups: Iterable[Callable[[Any], Execution]],
        n_runs: int | Iterable[int] = 11,
        n_threads: int = Parallelism.ACCURATE_TIME_MEASUREMENTS,
        perform_warmup: bool = True, warmup_fes: int = 20,
        perform_pre_warmup: bool = True, pre_warmup_fes: int = 20,
        on_completion: Callable[[Any, Path, Process], None] = __no_complete) \
        -> Path:
    """
    Run an experiment and store the log files into the given folder.

    This function will automatically run an experiment, i.e., apply a set
    `setups` of algorithm setups to a set `instances` of problem instances for
    `n_runs` each. It will collect log files and store them into an
    appropriate folder structure under the path `base_dir`. It will
    automatically draw random seeds for all algorithm runs using
    :func:`moptipy.utils.nputils.rand_seeds_from_str` based on the names of
    the problem instances to solve. This yields replicable experiments, i.e.,
    running the experiment program twice will yield exactly the same runs in
    exactly the same file structure (give and take clock-time dependent
    issues, which obviously cannot be controlled in a deterministic fashion).

    This function will use `n_threads` separate processes to parallelize the
    whole experiment (if you do not specify `n_threads`, it will be chosen
    automatically).

    Note for Windows users: The parallelization will not work under Windows.
    However, you can achieve *almost* the same effect and performance as for
    `n_threads=N` if you set `n_threads=1` and simply start the program `N`
    times separately (in separate terminals and in parallel). Of course, all
    `N` processes must have the same `base_dir` parameter. They will then
    automatically share the workload.

    :param base_dir: the base directory where to store the results
    :param instances: an iterable of callables, each of which should return an
        object representing a problem instance, whose `__str__` representation
        is a valid name
    :param setups: an iterable of callables, each receiving an instance (as
        returned by instances) as input and producing an
        :class:`moptipy.api.execution.Execution` as output
    :param n_runs: the number of runs per algorithm-instance combination
    :param n_threads: the number of parallel threads of execution to use.
        This parameter only works under Linux! It should be set to 1 under all
        other operating systems. Under Linux, by default, we will use the
        number of physical cores - 1 processes.
        The default value for `n_threads` is computed in \
:py:const:`~moptipy.api.experiment.Parallelism.ACCURATE_TIME_MEASUREMENTS`,
        which will be different for different machines(!).
        We will try to distribute the threads over different logical and
        physical cores to minimize their interactions. If n_threads is less
        or equal the number of physical cores, then multiple logical cores
        will be assigned to each process.
        If less threads than the number of physical cores are spawned, we will
        leave one physical core unoccupied. This core may be used by the
        operating system or other processes for their work, thus reducing
        interference of the os with our experiments.
    :param perform_warmup: should we perform a warm-up for each instance?
        If this parameter is `True`, then before the very first run of a
        thread on an instance, we will execute the algorithm for just a few
        function evaluations without logging and discard the results. The
        idea is that during this warm-up, things such as JIT compilation or
        complicated parsing can take place. While this cannot mitigate time
        measurement problems for JIT compilations taking place late in runs,
        it can at least somewhat solve the problem of delayed first FEs caused
        by compilation and parsing.
    :param warmup_fes: the number of the FEs for the warm-up runs
    :param perform_pre_warmup: should we do one warmup run for each
        instance before we begin with the actual experiments? This complements
        the warmups defined by `perform_warmup`. It could be that, for some
        reason, JIT or other activities may lead to stalls between multiple
        processes when code is encountered for the first time. This may or may
        not still cause strange timing issues even if `perform_warmup=True`.
        We therefore can do one complete round of warmups before starting the
        actual experiment. After that, we perform one garbage collection run
        and then freeze all objects surviving it to prevent them from future
        garbage collection runs. All processes that execute the experiment in
        parallel will complete their pre-warmup and only after all of them have
        completed it, the actual experiment will begin. I am not sure whether
        this makes sense or not, but it also would not hurt.
    :param pre_warmup_fes: the FEs for the pre-warmup runs
    :param on_completion: a function to be called for every completed run,
        receiving the instance, the path to the log file (before it is
        created) and the :class:`~moptipy.api.process.Process` of the run
        as parameters

    :returns: the canonicalized path to `base_dir`
    """
    if not isinstance(instances, Iterable):
        raise type_error(instances, "instances", Iterable)
    if not isinstance(setups, Iterable):
        raise type_error(setups, "setups", Iterable)
    if not isinstance(perform_warmup, bool):
        raise type_error(perform_warmup, "perform_warmup", bool)
    if not isinstance(perform_pre_warmup, bool):
        raise type_error(perform_pre_warmup, "perform_pre_warmup", bool)
    check_int_range(warmup_fes, "warmup_fes", 1, 1_000_000)
    check_int_range(pre_warmup_fes, "pre_warmup_fes", 1, 1_000_000)
    check_int_range(n_threads, "n_threads", 1, 16384)
    instances = list(instances)
    if len(instances) <= 0:
        raise ValueError("Instance enumeration is empty.")
    for instance in instances:
        if not callable(instance):
            raise type_error(instance, "all instances", call=True)

    sysinfo_check: str = get_sys_info()
    if not isinstance(sysinfo_check, str):
        raise type_error(sysinfo_check, "system information", str)
    if len(sysinfo_check) <= 0:
        raise ValueError(f"invalid system info {sysinfo_check!r}!")

    setups = list(setups)
    if len(setups) <= 0:
        raise ValueError("Setup enumeration is empty.")
    for setup in setups:
        if not callable(setup):
            raise type_error(setup, "all setups", call=True)

    experiments: Final[list[list[Callable]]] = \
        [[ii, ss] for ii in instances for ss in setups]

    del instances
    del setups

    if len(experiments) <= 0:
        raise ValueError("No experiments found?")

    n_runs = [n_runs] if isinstance(n_runs, int) else list(n_runs)
    last = 0
    for run in n_runs:
        last = check_int_range(run, "n_runs", last + 1)

    cache: Final[Callable[[str], bool]] = str_is_new()
    use_dir: Final[Path] = Path(base_dir)
    use_dir.ensure_dir_exists()

    stdio_lock: AbstractContextManager

    if n_threads > 1:
        file_lock: AbstractContextManager = mp.Lock()
        stdio_lock = mp.Lock()
        logger(f"starting experiment with {n_threads} threads "
               f"on {_CPU_LOGICAL_CORES} logical cores, "
               f"{_CPU_PHYSICAL_CORES} physical cores (i.e.,"
               f" {_CPU_LOGICAL_PER_PHYSICAL} logical cores per physical "
               "core).", "", stdio_lock)

        event: Final = mp.Event()
        pre_warmup_barrier: Final = mp.Barrier(n_threads) \
            if perform_pre_warmup else None
        processes: Final[list[mp.Process]] = \
            [mp.Process(target=__waiting_run_experiment,
                        args=(use_dir,
                              experiments.copy(),
                              n_runs,
                              perform_warmup,
                              warmup_fes,
                              perform_pre_warmup,
                              pre_warmup_fes,
                              file_lock,
                              stdio_lock,
                              cache,
                              ":" + hex(i)[2:],
                              event,
                              pre_warmup_barrier,
                              on_completion))
             for i in range(n_threads)]

        for i, p in enumerate(processes):
            p.start()
            logger(f"started processes {hex(i)[2:]} in waiting state.",
                   "", stdio_lock)

        # try to distribute the load evenly over all cores
        n_cpus: int = _CPU_PHYSICAL_CORES
        core_ofs: int = 0
        if n_threads < n_cpus:
            n_cpus -= 1
            core_ofs = _CPU_LOGICAL_PER_PHYSICAL
        n_cores: Final[int] = n_cpus * _CPU_LOGICAL_PER_PHYSICAL
        n_cores_per_thread: Final[int] = max(1, n_cores // n_threads)

        last_core: int = 0
        for i, p in enumerate(processes):
            pid: int = int(p.pid)
            aff: list[int] = []
            for _ in range(n_cores_per_thread):
                aff.append(int((last_core + core_ofs) % _CPU_LOGICAL_CORES))
                last_core += 1
            psutil.Process(pid).cpu_affinity(aff)
            logger(f"set affinity of processes {hex(i)[2:]} with "
                   f"pid {pid} ({hex(pid)}) to {aff}.", "", stdio_lock)
        logger("now releasing lock and starting all processes.",
               "", stdio_lock)
        event.set()
        for i, p in enumerate(processes):
            p.join()
            logger(f"processes {hex(i)[2:]} has finished.", "", stdio_lock)

    else:
        logger(f"starting experiment with single thread "
               f"on {_CPU_LOGICAL_CORES} logical cores, "
               f"{_CPU_PHYSICAL_CORES} physical cores (i.e.,"
               f" {_CPU_LOGICAL_PER_PHYSICAL} logical cores per physical "
               "core).")
        stdio_lock = nullcontext()
        __run_experiment(base_dir=use_dir,
                         experiments=experiments,
                         n_runs=n_runs,
                         perform_warmup=perform_warmup,
                         warmup_fes=warmup_fes,
                         perform_pre_warmup=perform_pre_warmup,
                         pre_warmup_fes=pre_warmup_fes,
                         file_lock=nullcontext(),
                         stdio_lock=stdio_lock,
                         cache=cache,
                         thread_id="",
                         pre_warmup_barrier=None,
                         on_completion=on_completion)

    logger("finished experiment.", "", stdio_lock)
    return use_dir



"""Evaluate the results of the example experiment."""
import argparse
import os.path as pp
from math import log2
from re import Match, Pattern
from re import compile as _compile
from statistics import median
from typing import Callable, Final, Iterable, cast

from pycommons.io.console import logger
from pycommons.io.path import Path, directory_path
from pycommons.strings.string_conv import num_to_str
from pycommons.types import type_error

from moptipy.evaluation.axis_ranger import AxisRanger
from moptipy.evaluation.base import TIME_UNIT_FES, TIME_UNIT_MILLIS
from moptipy.evaluation.end_results import EndResult
from moptipy.evaluation.end_results import from_csv as end_results_from_csv
from moptipy.evaluation.end_results import from_logs as end_results_from_logs
from moptipy.evaluation.end_results import to_csv as end_results_to_csv
from moptipy.evaluation.end_statistics import EndStatistics
from moptipy.evaluation.end_statistics import (
    from_end_results as es_from_end_results,
)
from moptipy.evaluation.end_statistics import to_csv as es_to_csv
from moptipy.evaluation.tabulate_end_results import (
    DEFAULT_ALGORITHM_INSTANCE_STATISTICS,
    DEFAULT_ALGORITHM_SUMMARY_STATISTICS,
    command_column_namer,
    tabulate_end_results,
)
from moptipy.evaluation.tabulate_result_tests import tabulate_result_tests
from moptipy.examples.jssp.experiment import ALGORITHMS, INSTANCES
from moptipy.examples.jssp.instance import Instance
from moptipy.examples.jssp.plots import (
    plot_end_makespans,
    plot_end_makespans_over_param,
    plot_progresses,
    plot_stat_gantt_charts,
)
from moptipy.spaces.permutations import Permutations
from moptipy.utils.help import moptipy_argparser
from moptipy.utils.lang import EN
from moptipy.utils.logger import sanitize_name
from moptipy.utils.strings import (
    beautify_float_str,
    name_str_to_num,
)

#: The letter mu
LETTER_M: Final[str] = "\u03BC"
#: The letter lambda
LETTER_L: Final[str] = "\u03BB"

#: the pre-defined instance sort keys
__INST_SORT_KEYS: Final[dict[str, int]] = {
    __n: __i for __i, __n in enumerate(INSTANCES)
}

#: the name of the mu+1 EA
NAME_EA_MU_PLUS_1: Final[str] = f"{LETTER_M}+1_ea"
#: the name of the mu+mu EA
NAME_EA_MU_PLUS_MU: Final[str] = f"{LETTER_M}+{LETTER_M}_ea"
#: the name of the mu+ld(mu) EA
NAME_EA_MU_PLUS_LOG_MU: Final[str] = f"{LETTER_M}+log\u2082{LETTER_M}_ea"
#: the name of the mu+sqrt(mu) EA
NAME_EA_MU_PLUS_SQRT_MU: Final[str] = f"{LETTER_M}+\u221A{LETTER_M}_ea"


def __make_all_names() -> tuple[str, ...]:
    """
    Create an immutable list of all algorithm names.

    :returns: the tuple with all algorithm names
    """
    inst = Instance.from_resource("demo")
    space = Permutations.with_repetitions(inst.jobs, inst.machines)
    return tuple([str(a(inst, space)) for a in ALGORITHMS])


#: the list of all algorithm names from the experiment
ALL_NAMES: Final[tuple[str, ...]] = __make_all_names()
del __make_all_names


def ea_family(name: str) -> str:
    """
    Name the evolutionary algorithm setup.

    :param name: the full name
    :returns: the short name of the family
    """
    if name.startswith("ea_"):
        ss: Final[list[str]] = name.split("_")
        ll: Final[int] = len(ss)
        if ll == 4:
            lambda_ = int(ss[2])
            if lambda_ == 1:
                return NAME_EA_MU_PLUS_1
            mu = int(ss[1])
            if lambda_ == int(log2(mu)):
                return NAME_EA_MU_PLUS_LOG_MU
            if lambda_ == round(mu ** 0.5):
                return NAME_EA_MU_PLUS_SQRT_MU
            ratio = lambda_ // mu
            if ratio * mu == lambda_:
                if ratio == 1:
                    return NAME_EA_MU_PLUS_MU
                return f"{LETTER_M}+{ratio}{LETTER_M}_ea"
        elif (ll == 6) and (ss[-2] == "gap") and (ss[-1] == "swap2"):
            return f"{ss[1]}+{ss[2]}_ea_br"
    raise ValueError(f"Invalid EA name {name!r}.")


def sa_family(name: str) -> str:
    """
    Name the simulated annealing setup.

    :param name: the full name
    :returns: the short name of the family
    """
    if name.startswith("sa_") and name.endswith("_swap2"):
        return (f"sa_T\u2080_"
                f"{beautify_float_str(name_str_to_num(name.split('_')[2]))}")
    raise ValueError(f"Invalid SA name {name!r}.")


def ma_family(name: str) -> str:
    """
    Compute the Memetic Algorithm family without the ls steps.

    :param name: the algorithm name
    :returns: the short name of the family
    """
    ss: Final[list[str]] = name.split("_")
    s = "rls" if name.startswith("marls") else "sa"
    return f"{ss[1]}+{ss[2]}_ma{s}"


def __make_algo_names() -> tuple[dict[str, int], dict[str, str]]:
    """
    Create the algorithm sort keys and name decode.

    :returns: the algorithm sort keys and name decode
    """
    names: list[str] = list(ALL_NAMES)
    names_new = list(ALL_NAMES)

    def __eacr(mtch: Match) -> str:
        """Transform the EA name with crossover."""
        br = name_str_to_num(mtch.group(3))
        if not 0.0 < br < 1.0:
            raise ValueError(f"{mtch} has invalid br={br}.")
        a = log2(br)
        b = int(a)
        c = log2(1.0 - br)
        d = int(c)
        q = f"(1-2^{d})" if abs(c - d) < abs(a - b) else f"2^{b}"
        q2 = num_to_str(br)
        if len(q2) <= (len(q) + 1):
            q = q2
        return f"{mtch.group(1)}+{mtch.group(2)}_ea_{q}"

    def __sa(mtch: Match) -> str:
        """Transform the SA name."""
        if mtch.group(0) == "sa_exp16_1em6_swap2":
            return "sa"
        return f"sa_{mtch.group(1)}_{mtch.group(2)}e-{mtch.group(3)}"

    def __ma(mtch: Match, suffix: str) -> str:
        """Transform the MA name."""
        if mtch.group(0) == "ma_2_2_1048576_gap_sa_exp16_5d1em6_swap2":
            return "2_masa_20"
        if mtch.group(0) == "marls_8_8_16_gap_swap2":
            return "8_marls_4"
        ls = name_str_to_num(mtch.group(3))
        if not 0 < ls < 1_000_000_000:
            raise ValueError(f"{mtch} has invalid ls={ls}.")
        e2 = int(log2(ls))
        if ls != (2 ** e2):
            raise ValueError(
                f"{mtch} has invalid ls={ls} since {e2}=log2({ls}).")
        e2s = f"2^{e2}"
        lss = str(ls)
        if len(lss) <= len(e2s):
            e2s = lss
        return f"{mtch.group(1)}+{mtch.group(2)}_ma{suffix}_{e2s}"

    # fix some names using regular expressions
    namer: dict[str, str] = {}
    used_names: set[str] = set(names)
    for pattern, repl in [("hc_swap2", "hc"), ("hc_swapn", "hcn"),
                          ("rls_swap2", "rls"), ("rls_swapn", "rlsn")]:
        re: Pattern = _compile(pattern)
        found = False
        for s in names:
            m: Match = re.match(s)
            if m is not None:
                ns = re.sub(cast(str | Callable[[Match], str], repl), s)
                if (ns is None) or (len(ns) <= 0):
                    raise ValueError(f"{s!r} -> {ns!r}?")
                if ns == s:
                    continue
                found = True
                os = namer.get(s)
                if os is not None:
                    if os == ns:
                        continue
                    raise ValueError(f"{s!r} -> {ns!r}, {os!r}?")
                if ns in used_names:
                    raise ValueError(f"Already got {ns!r}.")
                namer[s] = ns
                names_new.insert(names_new.index(s), ns)
        if not found:
            raise ValueError(f"did not find {pattern!r}.")

    return {__n: __i for __i, __n in enumerate(names_new)}, namer


#: the pre-defined algorithm sort keys and name map
__ALGO_SORT_KEYS, __ALGO_NAME_MAP = __make_algo_names()


def instance_sort_key(name: str) -> int:
    """
    Get the instance sort key for a given instance name.

    :param name: the instance name
    :returns: the sort key
    """
    if not isinstance(name, str):
        raise type_error(name, "name", str)
    if not name:
        raise ValueError("name must not be empty.")
    if name in __INST_SORT_KEYS:
        return __INST_SORT_KEYS[name]
    return 1000


def algorithm_sort_key(name: str) -> int:
    """
    Get the algorithm sort key for a given algorithm name.

    :param name: the algorithm name
    :returns: the sort key
    """
    if not isinstance(name, str):
        raise type_error(name, "name", str)
    if not name:
        raise ValueError("name must not be empty.")
    if name in __ALGO_SORT_KEYS:
        return __ALGO_SORT_KEYS[name]
    return 1000


def algorithm_namer(name: str) -> str:
    """
    Rename algorithm setups.

    :param name: the algorithm's original name
    :returns: the new name of the algorithm
    """
    if not isinstance(name, str):
        raise type_error(name, "name", str)
    if not name:
        raise ValueError("name must not be empty.")
    if name in __ALGO_NAME_MAP:
        return __ALGO_NAME_MAP[name]
    return name


def compute_end_results(results_dir: str,
                        dest_dir: str) -> Path:
    """
    Get the end results, compute them if necessary.

    :param results_dir: the results directory
    :param dest_dir: the destination directory
    :returns: the path to the end results file.
    """
    dest: Final[Path] = directory_path(dest_dir)
    results_file: Final[Path] = dest.resolve_inside("end_results.txt")
    if results_file.is_file():
        return results_file
    results: Final[list[EndResult]] = []

    source: Final[Path] = directory_path(results_dir)
    logger(f"loading end results from path {source!r}.")
    end_results_from_logs(source, results.append)
    if not results:
        raise ValueError(f"Could not find any logs in {source!r}.")
    results.sort()
    logger(f"found {len(results)} log files in path {source!r}, storing "
           f"them in file {results_file!r}.")
    rf: Path = end_results_to_csv(results, results_file)
    if rf != results_file:
        raise ValueError(
            f"results file should be {results_file!r}, but is {rf!r}.")
    results_file.enforce_file()
    logger(f"finished writing file {results_file!r}.")
    return results_file


def get_end_results(
        file: str,
        insts: None | set[str] | Callable[[str], bool] = None,
        algos: None | set[str] | Callable[[str], bool] = None) \
        -> list[EndResult]:
    """
    Get a specific set of end results..

    :param file: the end results file
    :param insts: only these instances will be included if this parameter is
        provided
    :param algos: only these algorithms will be included if this parameter is
        provided
    """

    def __filter(er: EndResult,
                 ins=None if insts is None else
                 insts.__contains__ if isinstance(insts, set)
                 else insts,
                 alg=None if algos is None else
                 algos.__contains__ if isinstance(algos, set)
                 else algos) -> bool:
        if (ins is not None) and (not ins(er.instance)):
            return False
        if (alg is not None) and (not alg(er.algorithm)):
            return False
        return True

    col: Final[list[EndResult]] = []
    end_results_from_csv(file=file, consumer=col.append, filterer=__filter)
    if len(col) <= 0:
        raise ValueError(
            f"no end results for instances {insts} and algorithms {algos}.")
    return col


def compute_end_statistics(end_results_file: str,
                           dest_dir: str) -> Path:
    """
    Get the end result statistics, compute them if necessary.

    :param end_results_file: the end results file
    :param dest_dir: the destination directory
    :returns: the path to the end result statistics file.
    """
    dest: Final[Path] = directory_path(dest_dir)
    stats_file: Final[Path] = dest.resolve_inside("end_statistics.txt")
    if stats_file.is_file():
        return stats_file

    results: Final[list[EndResult]] = get_end_results(end_results_file)
    if len(results) <= 0:
        raise ValueError("end results cannot be empty")
    stats: Final[list[EndStatistics]] = []
    es_from_end_results(results, stats.append)
    if len(stats) <= 0:
        raise ValueError("end result statistics cannot be empty")
    stats.sort()

    sf: Path = es_to_csv(stats, stats_file)
    if sf != stats_file:
        raise ValueError(f"stats file should be {stats_file!r} but is {sf!r}")
    stats_file.enforce_file()
    logger(f"finished writing file {stats_file!r}.")
    return stats_file


def table(end_results: Path, algos: list[str], dest: Path,
          swap_stats: Iterable[tuple[str, str]] | None = None) -> None:
    """
    Tabulate the end results.

    :param end_results: the path to the end results
    :param algos: the algorithms
    :param dest: the directory
    :param swap_stats: the statistics to swap out
    """
    EN.set_current()
    n: Final[str] = algorithm_namer(algos[0])
    algo_inst_stat: list | tuple = DEFAULT_ALGORITHM_INSTANCE_STATISTICS
    algo_sum_stat: list | tuple = DEFAULT_ALGORITHM_SUMMARY_STATISTICS
    if swap_stats is not None:
        algo_inst_stat = list(algo_inst_stat)
        algo_sum_stat = list(algo_sum_stat)
        for old, swap in swap_stats:
            found: bool = False
            for lst in [algo_inst_stat, algo_sum_stat]:
                for idx, elem in enumerate(lst):
                    if elem == old:
                        found = True
                        lst[idx] = swap
                        break
            if not found:
                raise ValueError(f"did not find {old!r} in {str(lst)!r}.")

    tabulate_end_results(
        end_results=get_end_results(end_results, algos=set(algos)),
        file_name=sanitize_name(f"end_results_{n}"), dir_name=dest,
        instance_sort_key=instance_sort_key,
        algorithm_sort_key=algorithm_sort_key,
        col_namer=command_column_namer,
        algorithm_namer=algorithm_namer,
        algorithm_instance_statistics=algo_inst_stat,
        algorithm_summary_statistics=algo_sum_stat,
        use_lang=False)


def tests(end_results: Path, algos: list[str], dest: Path) -> None:
    """
    Tabulate the end result tests.

    :param end_results: the path to the end results
    :param algos: the algorithms
    :param dest: the directory
    """
    EN.set_current()
    n: Final[str] = algorithm_namer(algos[0])
    tabulate_result_tests(
        end_results=get_end_results(end_results, algos=set(algos)),
        file_name=sanitize_name(f"tests_{n}"), dir_name=dest,
        instance_sort_key=instance_sort_key,
        algorithm_sort_key=algorithm_sort_key,
        algorithm_namer=algorithm_namer,
        use_lang=False)


def makespans(end_results: Path, algos: list[str], dest: Path,
              x_label_location: float = 1.0) -> None:
    """
    Plot the end makespans.

    :param end_results: the path to the end results
    :param algos: the algorithms
    :param dest: the directory
    :param x_label_location: the location of the label of the x-axis
    """
    n: Final[str] = algorithm_namer(algos[0])
    plot_end_makespans(
        end_results=get_end_results(end_results, algos=set(algos)),
        name_base=sanitize_name(f"makespan_scaled_{n}"), dest_dir=dest,
        instance_sort_key=instance_sort_key,
        algorithm_sort_key=algorithm_sort_key,
        algorithm_namer=algorithm_namer,
        x_label_location=x_label_location)


def gantt(end_results: Path, algo: str, dest: Path, source: Path,
          best: bool = False,
          insts: Iterable[str] | None = None) -> None:
    """
    Plot the median Gantt charts.

    :param end_results: the path to the end results
    :param algo: the algorithm
    :param dest: the directory
    :param source: the source directory
    :param best: should we plot the best instance only (or, otherwise, the
        median)
    :param insts: the instances to use
    """
    n: Final[str] = algorithm_namer(algo)
    plot_stat_gantt_charts(
        get_end_results(end_results, algos={algo},
                        insts=None if insts is None else set(insts)),
        name_base=sanitize_name(f"best_gantt_{n}" if best else f"gantt_{n}"),
        dest_dir=dest,
        results_dir=source,
        instance_sort_key=instance_sort_key,
        statistic=cast(
            Callable[[Iterable[int | float]], int | float],
            min if best else median))


def progress(algos: list[str], dest: Path, source: Path,
             log: bool = True, millis: bool = True) -> None:
    """
    Plot the median Gantt charts.

    :param algos: the algorithms
    :param dest: the directory
    :param source: the source directory
    :param log: is the time logarithmically scaled?
    :param millis: is the time measured in milliseconds?
    """
    n: str = f"progress_{algorithm_namer(algos[0])}_"
    if log:
        n = n + "log_"
    if millis:
        unit = TIME_UNIT_MILLIS
        n = n + "T"
    else:
        unit = TIME_UNIT_FES
        n = n + "FEs"
    plot_progresses(results_dir=source,
                    algorithms=algos,
                    name_base=n,
                    dest_dir=dest,
                    log_time=log,
                    time_unit=unit,
                    algorithm_namer=algorithm_namer,
                    instance_sort_key=instance_sort_key,
                    algorithm_sort_key=algorithm_sort_key)


def makespans_over_param(
        end_results: Path,
        selector: Callable[[str], bool],
        x_getter: Callable[[EndStatistics], int | float],
        name_base: str,
        dest_dir: str,
        x_axis: AxisRanger | Callable[[], AxisRanger]
        = AxisRanger,
        x_label: str | None = None,
        algo_getter: Callable[[str], str] | None = None,
        title: str | None = None,
        legend_pos: str = "right",
        title_x: float = 0.5,
        y_label_location: float = 1.0) -> list[Path]:
    """
    Plot the performance over a parameter.

    :param end_results: the end results path
    :param selector: the selector for algorithms
    :param name_base: the basic name
    :param dest_dir: the destination directory
    :param x_getter: the function computing the x-value for each statistics
        object
    :param x_axis: the axis ranger
    :param x_label: the x-axis label
    :param algo_getter: the optional algorithm name getter (use `name_base` if
        unspecified)
    :param title: the optional title (use `name_base` if unspecified)
    :param legend_pos: the legend position, set to "right"
    :param title_x: the title x
    :param y_label_location: the y label location
    :returns: the list of generated files
    """

    def _algo_name_getter(es: EndStatistics,
                          n=name_base, g=algo_getter) -> str:
        return n if g is None else g(es.algorithm)

    return plot_end_makespans_over_param(
        end_results=get_end_results(end_results, algos=selector),
        x_getter=x_getter, name_base=sanitize_name(f"{name_base}_results"),
        dest_dir=dest_dir, title=name_base if title is None else title,
        algorithm_getter=_algo_name_getter,  # type: ignore
        instance_sort_key=instance_sort_key,
        algorithm_sort_key=algorithm_sort_key,
        x_axis=x_axis, x_label=x_label,
        plot_single_instances=algo_getter is None,
        legend_pos=legend_pos, title_x=title_x,
        y_label_location=y_label_location)


def evaluate_experiment(results_dir: str = pp.join(".", "results"),
                        dest_dir: str | None = None) -> None:
    """
    Evaluate the experiment.

    :param results_dir: the results directory
    :param dest_dir: the destination directory
    """
    source: Final[Path] = directory_path(results_dir)
    dest: Final[Path] = Path(dest_dir if dest_dir else
                             pp.join(source, "..", "evaluation"))
    dest.ensure_dir_exists()
    logger(f"Beginning evaluation from {source!r} to {dest!r}.")

    end_results: Final[Path] = compute_end_results(source, dest)
    if not end_results:
        raise ValueError("End results path is empty??")
    end_stats: Final[Path] = compute_end_statistics(end_results, dest)
    if not end_stats:
        raise ValueError("End stats path is empty??")

    logger("Now evaluating the single random sampling algorithm `1rs`.")
    table(end_results, ["1rs"], dest)
    makespans(end_results, ["1rs"], dest)
    gantt(end_results, "1rs", dest, source)

    logger("Now evaluating the multi-random sampling algorithm `rs`.")
    table(end_results, ["rs", "1rs"], dest)
    makespans(end_results, ["rs", "1rs"], dest)
    gantt(end_results, "rs", dest, source)
    progress(["rs"], dest, source)
    progress(["rs"], dest, source, log=False)

    logger("Now evaluating the hill climbing algorithm `hc`.")
    table(end_results, ["hc_swap2", "rs"], dest)
    makespans(end_results, ["hc_swap2", "rs"], dest)
    gantt(end_results, "hc_swap2", dest, source)
    progress(["hc_swap2", "rs"], dest, source)
    progress(["hc_swap2", "rs"], dest, source, millis=False)

    logger(f"Finished evaluation from {source!r} to {dest!r}.")


# Evaluate experiment if run as script
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = moptipy_argparser(
        __file__, "Evaluate the results of the JSSP example experiment",
        "This experiment evaluates all the results of the JSSP example"
        " experiment and creates the figures and tables of the "
        "'Optimization Algorithms' book (see "
        "http://thomasweise.github.io/oa).")
    parser.add_argument(
        "source", nargs="?", default="./results", type=Path,
        help="the directory with the results of the JSSP experiment")
    parser.add_argument(
        "dest", type=Path, nargs="?", default="./evaluation/",
        help="the directory to write the evaluation results to")
    args: Final[argparse.Namespace] = parser.parse_args()
    evaluate_experiment(results_dir=args.source, dest_dir=args.dest)



"""Some internal helper functions and base classes."""

from dataclasses import dataclass
from typing import Any, Callable, Final

from pycommons.types import check_int_range, type_error

from moptipy.utils.nputils import rand_seed_check
from moptipy.utils.strings import sanitize_name
from moptipy.version import __version__ as moptipy_version

#: The key for the total number of runs.
KEY_N: Final[str] = "n"
#: a key for the objective function name
KEY_OBJECTIVE_FUNCTION: Final[str] = "objective"
#: a key for the encoding name
KEY_ENCODING: Final[str] = "encoding"

#: The unit of the time axis if time is measured in milliseconds.
TIME_UNIT_MILLIS: Final[str] = "ms"
#: The unit of the time axis of time is measured in FEs
TIME_UNIT_FES: Final[str] = "FEs"

#: The name of the raw objective values data.
F_NAME_RAW: Final[str] = "plainF"
#: The name of the scaled objective values data.
F_NAME_SCALED: Final[str] = "scaledF"
#: The name of the normalized objective values data.
F_NAME_NORMALIZED: Final[str] = "normalizedF"


def check_time_unit(time_unit: Any) -> str:
    """
    Check that the time unit is OK.

    :param time_unit: the time unit
    :return: the time unit string

    >>> check_time_unit("FEs")
    'FEs'
    >>> check_time_unit("ms")
    'ms'
    >>> try:
    ...     check_time_unit(1)
    ... except TypeError as te:
    ...     print(te)
    time_unit should be an instance of str but is int, namely '1'.
    >>> try:
    ...     check_time_unit("blabedibla")
    ... except ValueError as ve:
    ...     print(ve)
    Invalid time unit 'blabedibla', only 'FEs' and 'ms' are permitted.
    """
    if not isinstance(time_unit, str):
        raise type_error(time_unit, "time_unit", str)
    if time_unit in (TIME_UNIT_FES, TIME_UNIT_MILLIS):
        return time_unit
    raise ValueError(
        f"Invalid time unit {time_unit!r}, only {TIME_UNIT_FES!r} "
        f"and {TIME_UNIT_MILLIS!r} are permitted.")


def check_f_name(f_name: Any) -> str:
    """
    Check whether an objective value name is valid.

    :param f_name: the name of the objective function dimension
    :return: the name of the objective function dimension

    >>> check_f_name("plainF")
    'plainF'
    >>> check_f_name("scaledF")
    'scaledF'
    >>> check_f_name("normalizedF")
    'normalizedF'
    >>> try:
    ...     check_f_name(1.0)
    ... except TypeError as te:
    ...     print(te)
    f_name should be an instance of str but is float, namely '1.0'.
    >>> try:
    ...     check_f_name("oops")
    ... except ValueError as ve:
    ...     print(ve)
    Invalid f name 'oops', only 'plainF', 'scaledF', and 'normalizedF' \
are permitted.
    """
    if not isinstance(f_name, str):
        raise type_error(f_name, "f_name", str)
    if f_name in (F_NAME_RAW, F_NAME_SCALED, F_NAME_NORMALIZED):
        return f_name
    raise ValueError(
        f"Invalid f name {f_name!r}, only {F_NAME_RAW!r}, "
        f"{F_NAME_SCALED!r}, and {F_NAME_NORMALIZED!r} are permitted.")


def _set_name(dest: object, name: str, what: str,
              none_allowed: bool = False,
              empty_to_none: bool = True) -> None:
    """
    Check and set a name.

    :param dest: the destination
    :param name: the name to set
    :param what: the name's type
    :param none_allowed: is `None` allowed?
    :param empty_to_none: If both `none_allowed` and `empty_to_none` are
        `True`, then empty strings are converted to `None`

    >>> class TV:
    ...     algorithm: str
    ...     instance: str | None
    >>> t = TV()
    >>> _set_name(t, "bla", "algorithm", False)
    >>> t.algorithm
    'bla'
    >>> _set_name(t, "xbla", "instance", True)
    >>> t.instance
    'xbla'
    >>> _set_name(t, None, "instance", True)
    >>> print(t.instance)
    None
    >>> t.instance = "x"
    >>> _set_name(t, "  ", "instance", True)
    >>> print(t.instance)
    None
    >>> try:
    ...     _set_name(t, 1, "algorithm")
    ... except TypeError as te:
    ...     print(te)
    algorithm name should be an instance of str but is int, namely '1'.
    >>> t.algorithm
    'bla'
    >>> try:
    ...     _set_name(t, "  ", "algorithm")
    ... except ValueError as ve:
    ...     print(ve)
    algorithm name cannot be empty of just consist of white space, but \
'  ' does.
    >>> t.algorithm
    'bla'
    >>> try:
    ...     _set_name(t, "a a", "instance")
    ... except ValueError as ve:
    ...     print(ve)
    Invalid instance name 'a a'.
    >>> print(t.instance)
    None
    >>> try:
    ...     _set_name(t, " ", "instance", True, False)
    ... except ValueError as ve:
    ...     print(ve)
    instance name cannot be empty of just consist of white space, but \
' ' does.
    >>> print(t.instance)
    None
    """
    use_name = name
    if isinstance(name, str):
        use_name = use_name.strip()
        if len(use_name) <= 0:
            if empty_to_none and none_allowed:
                use_name = None
            else:
                raise ValueError(f"{what} name cannot be empty of just cons"
                                 f"ist of white space, but {name!r} does.")
        elif use_name != sanitize_name(use_name):
            raise ValueError(f"Invalid {what} name {name!r}.")
    elif not ((name is None) and none_allowed):
        raise type_error(name, f"{what} name",
                         (str, None) if none_allowed else str)
    object.__setattr__(dest, what, use_name)


class EvaluationDataElement:
    """A base class for all the data classes in this module."""

    def _tuple(self) -> tuple[Any, ...]:
        """
        Create a tuple with all the data of this data class for comparison.

        All the relevant data of an instance of this class is stored in a
        tuple. The tuple is then used in the dunder methods for comparisons.
        The returned tuple *must* be based on the scheme
        `tuple[str, str, str, str, str, int, int, str, str]`.
        They can be shorter than this and they can be longer, but they must
        adhere to this basic scheme:

        1. class name
        2. algorithm name, or `""` if algorithm name is `None`
        3. instance name, or `""` if instance name is `None`
        4. objective name, `""` objective name is `None`
        5. encoding name, or `""` encoding name is `None`
        6. number of runs, or `0` if no number of runs is specified or `1` if
           the data concerns exactly one run
        7. the random seed, or `-1` if no random seed is specified
        8. the string time unit, or `""` if no time unit is given
        9. the scaling name of the objective function, or `""` if no scaling
           name is given

        If the tuples are longer, then all values following after this must be
        integers or floats.

        >>> EvaluationDataElement()._tuple()
        ('EvaluationDataElement',)

        :returns: a tuple with all the data of this class, where `None` values
            are masked out
        """
        return (self.__class__.__name__, )

    def __eq__(self, other) -> bool:
        """
        Compare for `==` with another object based on the `_tuple()` value.

        :param other: the other object to compare to, must be an instance of
            :class:`EvaluationDataElement`
        :retval `True`: if the `other` object's `_tuple()` representation is
            `==` with this object's `_tuple()` representation
        :retval `False`: otherwise
        :raises NotImplementedError: if the other object is not an instance of
            :class:`EvaluationDataElement` and therefore cannot be compared.

        >>> PerRunData("a", "i", "f", "e", 234) == PerRunData(
        ...     "a", "i", "f", "e", 234)
        True
        >>> PerRunData("a", "i", "f", "e", 234) == PerRunData(
        ...     "a", "j", "f", "e", 234)
        False
        >>> try:
        ...     PerRunData("a", "i", "f", "e", 234) == 3
        ... except NotImplementedError as ni:
        ...     print(ni)
        Cannot compare PerRunData(algorithm='a', instance='i', \
objective='f', encoding='e', rand_seed=234) with 3 for ==.
        """
        if isinstance(other, EvaluationDataElement):
            return self._tuple() == other._tuple()
        raise NotImplementedError(
            f"Cannot compare {self} with {other} for ==.")

    def __ne__(self, other) -> bool:
        """
        Compare for `!=` with another object based on the `_tuple()` value.

        :param other: the other object to compare to, must be an instance of
            :class:`EvaluationDataElement`
        :retval `True`: if the `other` object's `_tuple()` representation is
            `!=` with this object's `_tuple()` representation
        :retval `False`: otherwise
        :raises NotImplementedError: if the other object is not an instance of
            :class:`EvaluationDataElement` and therefore cannot be compared.

        >>> PerRunData("a", "i", "f", "e", 234) != PerRunData(
        ...     "a", "i", "f", "e", 234)
        False
        >>> PerRunData("a", "i", "f", "e", 234) != PerRunData(
        ...     "a", "j", "f", "e", 234)
        True
        >>> try:
        ...     PerRunData("a", "i", "f", "e", 234) != 3
        ... except NotImplementedError as ni:
        ...     print(ni)
        Cannot compare PerRunData(algorithm='a', instance='i', \
objective='f', encoding='e', rand_seed=234) with 3 for !=.
        """
        if isinstance(other, EvaluationDataElement):
            return self._tuple() != other._tuple()
        raise NotImplementedError(
            f"Cannot compare {self} with {other} for !=.")

    def __lt__(self, other) -> bool:
        """
        Compare for `<` with another object based on the `_tuple()` value.

        :param other: the other object to compare to, must be an instance of
            :class:`EvaluationDataElement`
        :retval `True`: if the `other` object's `_tuple()` representation is
            `<` with this object's `_tuple()` representation
        :retval `False`: otherwise
        :raises NotImplementedError: if the other object is not an instance of
            :class:`EvaluationDataElement` and therefore cannot be compared.

        >>> PerRunData("a", "i", "f", "e", 234) < PerRunData(
        ...     "a", "i", "f", "e", 234)
        False
        >>> PerRunData("a", "i", "f", "e", 234) < PerRunData(
        ...     "a", "j", "f", "e", 234)
        True
        >>> PerRunData("a", "j", "f", "e", 234) < PerRunData(
        ...     "a", "i", "f", "e", 234)
        False
        >>> try:
        ...     PerRunData("a", "i", "f", "e", 234) < 3
        ... except NotImplementedError as ni:
        ...     print(ni)
        Cannot compare PerRunData(algorithm='a', instance='i', \
objective='f', encoding='e', rand_seed=234) with 3 for <.
        """
        if isinstance(other, EvaluationDataElement):
            return self._tuple() < other._tuple()
        raise NotImplementedError(
            f"Cannot compare {self} with {other} for <.")

    def __le__(self, other) -> bool:
        """
        Compare for `<=` with another object based on the `_tuple()` value.

        :param other: the other object to compare to, must be an instance of
            :class:`EvaluationDataElement`
        :retval `True`: if the `other` object's `_tuple()` representation is
            `<=` with this object's `_tuple()` representation
        :retval `False`: otherwise
        :raises NotImplementedError: if the other object is not an instance of
            :class:`EvaluationDataElement` and therefore cannot be compared.

        >>> PerRunData("a", "i", "f", "e", 234) <= PerRunData(
        ...     "a", "i", "f", "e", 234)
        True
        >>> PerRunData("a", "i", "f", "e", 234) <= PerRunData(
        ...     "a", "j", "f", "e", 234)
        True
        >>> PerRunData("a", "j", "f", "e", 234) < PerRunData(
        ...     "a", "i", "f", "e", 234)
        False
        >>> try:
        ...     PerRunData("a", "i", "f", "e", 234) <= 3
        ... except NotImplementedError as ni:
        ...     print(ni)
        Cannot compare PerRunData(algorithm='a', instance='i', \
objective='f', encoding='e', rand_seed=234) with 3 for <=.
        """
        if isinstance(other, EvaluationDataElement):
            return self._tuple() <= other._tuple()
        raise NotImplementedError(
            f"Cannot compare {self} with {other} for <=.")

    def __gt__(self, other) -> bool:
        """
        Compare for `>` with another object based on the `_tuple()` value.

        :param other: the other object to compare to, must be an instance of
            :class:`EvaluationDataElement`
        :retval `True`: if the `other` object's `_tuple()` representation is
            `>` with this object's `_tuple()` representation
        :retval `False`: otherwise
        :raises NotImplementedError: if the other object is not an instance of
            :class:`EvaluationDataElement` and therefore cannot be compared.

        >>> PerRunData("a", "i", "f", "e", 234) > PerRunData(
        ...     "a", "i", "f", "e", 234)
        False
        >>> PerRunData("a", "i", "f", "e", 234) > PerRunData(
        ...     "a", "j", "f", "e", 234)
        False
        >>> PerRunData("a", "j", "f", "e", 234) > PerRunData(
        ...     "a", "i", "f", "e", 234)
        True
        >>> try:
        ...     PerRunData("a", "i", "f", "e", 234) > 3
        ... except NotImplementedError as ni:
        ...     print(ni)
        Cannot compare PerRunData(algorithm='a', instance='i', \
objective='f', encoding='e', rand_seed=234) with 3 for >.
        """
        if isinstance(other, EvaluationDataElement):
            return self._tuple() > other._tuple()
        raise NotImplementedError(
            f"Cannot compare {self} with {other} for >.")

    def __ge__(self, other) -> bool:
        """
        Compare for `>=` with another object based on the `_tuple()` value.

        :param other: the other object to compare to, must be an instance of
            :class:`EvaluationDataElement`
        :retval `True`: if the `other` object's `_tuple()` representation is
            `>=` with this object's `_tuple()` representation
        :retval `False`: otherwise
        :raises NotImplementedError: if the other object is not an instance of
            :class:`EvaluationDataElement` and therefore cannot be compared.

        >>> PerRunData("a", "i", "f", "e", 234) >= PerRunData(
        ...     "a", "i", "f", "e", 234)
        True
        >>> PerRunData("a", "i", "f", "e", 234) >= PerRunData(
        ...     "a", "j", "f", "e", 234)
        False
        >>> PerRunData("a", "j", "f", "e", 234) >= PerRunData(
        ...     "a", "i", "f", "e", 234)
        True
        >>> try:
        ...     PerRunData("a", "i", "f", "e", 234) >= 3
        ... except NotImplementedError as ni:
        ...     print(ni)
        Cannot compare PerRunData(algorithm='a', instance='i', \
objective='f', encoding='e', rand_seed=234) with 3 for >=.
        """
        if isinstance(other, EvaluationDataElement):
            return self._tuple() >= other._tuple()
        raise NotImplementedError(
            f"Cannot compare {self} with {other} for >=.")


@dataclass(frozen=True, init=False, order=False, eq=False)
class PerRunData(EvaluationDataElement):
    """
    An immutable record of information over a single run.

    >>> p = PerRunData("a", "i", "f", None, 234)
    >>> p.instance
    'i'
    >>> p.algorithm
    'a'
    >>> p.objective
    'f'
    >>> print(p.encoding)
    None
    >>> p.rand_seed
    234
    >>> p = PerRunData("a", "i", "f", "e", 234)
    >>> p.instance
    'i'
    >>> p.algorithm
    'a'
    >>> p.objective
    'f'
    >>> p.encoding
    'e'
    >>> p.rand_seed
    234
    >>> try:
    ...     PerRunData(3, "i", "f", "e", 234)
    ... except TypeError as te:
    ...     print(te)
    algorithm name should be an instance of str but is int, namely '3'.
    >>> try:
    ...     PerRunData("@1 2", "i", "f", "e", 234)
    ... except ValueError as ve:
    ...     print(ve)
    Invalid algorithm name '@1 2'.
    >>> try:
    ...     PerRunData("x", 3.2, "f", "e", 234)
    ... except TypeError as te:
    ...     print(te)
    instance name should be an instance of str but is float, namely '3.2'.
    >>> try:
    ...     PerRunData("x", "sdf i", "f", "e", 234)
    ... except ValueError as ve:
    ...     print(ve)
    Invalid instance name 'sdf i'.
    >>> try:
    ...     PerRunData("a", "i", True, "e", 234)
    ... except TypeError as te:
    ...     print(te)
    objective name should be an instance of str but is bool, namely 'True'.
    >>> try:
    ...     PerRunData("x", "i", "d-f", "e", 234)
    ... except ValueError as ve:
    ...     print(ve)
    Invalid objective name 'd-f'.
    >>> try:
    ...     PerRunData("x", "i", "f", 54.2, 234)
    ... except TypeError as te:
    ...     print(te)
    encoding name should be an instance of any in {None, str} but is float, \
namely '54.2'.
    >>> try:
    ...     PerRunData("y", "i", "f", "x  x", 234)
    ... except ValueError as ve:
    ...     print(ve)
    Invalid encoding name 'x  x'.
    >>> try:
    ...     PerRunData("x", "i", "f", "e", 3.3)
    ... except TypeError as te:
    ...     print(te)
    rand_seed should be an instance of int but is float, namely '3.3'.
    >>> try:
    ...     PerRunData("x", "i", "f", "e", -234)
    ... except ValueError as ve:
    ...     print(ve)
    rand_seed=-234 is invalid, must be in 0..18446744073709551615.
    """

    #: The algorithm that was applied.
    algorithm: str
    #: The problem instance that was solved.
    instance: str
    #: the name of the objective function
    objective: str
    #: the encoding, if any, or `None` if no encoding was used
    encoding: str | None
    #: The seed of the random number generator.
    rand_seed: int

    def __init__(self, algorithm: str, instance: str, objective: str,
                 encoding: str | None, rand_seed: int):
        """
        Create a per-run data record.

        :param algorithm: the algorithm name
        :param instance: the instance name
        :param objective: the name of the objective function
        :param encoding: the name of the encoding that was used, if any, or
            `None` if no encoding was used
        :param rand_seed: the random seed
        """
        _set_name(self, algorithm, "algorithm")
        _set_name(self, instance, "instance")
        _set_name(self, objective, "objective")
        _set_name(self, encoding, "encoding", True, False)
        object.__setattr__(self, "rand_seed", rand_seed_check(rand_seed))

    def _tuple(self) -> tuple[Any, ...]:
        """
        Get the tuple representation of this object used in comparisons.

        :return: the comparison-relevant data of this object in a tuple

        >>> PerRunData("a", "i", "f", "e", 234)._tuple()
        ('PerRunData', 'a', 'i', 'f', 'e', 1, 234)
        >>> PerRunData("a", "i", "f", None, 234)._tuple()
        ('PerRunData', 'a', 'i', 'f', '', 1, 234)
        """
        return (self.__class__.__name__, self.algorithm, self.instance,
                self.objective,
                "" if self.encoding is None else self.encoding, 1,
                self.rand_seed)


@dataclass(frozen=True, init=False, order=False, eq=False)
class MultiRunData(EvaluationDataElement):
    """
    A class that represents statistics over a set of runs.

    If one algorithm*instance is used, then `algorithm` and `instance` are
    defined. Otherwise, only the parameter which is the same over all recorded
    runs is defined.

    >>> p = MultiRunData("a", "i", "f", None, 3)
    >>> p.instance
    'i'
    >>> p.algorithm
    'a'
    >>> p.objective
    'f'
    >>> print(p.encoding)
    None
    >>> p.n
    3
    >>> p = MultiRunData(None, None, None, "x", 3)
    >>> print(p.instance)
    None
    >>> print(p.algorithm)
    None
    >>> print(p.objective)
    None
    >>> p.encoding
    'x'
    >>> p.n
    3
    >>> try:
    ...     MultiRunData(1, "i", "f", "e", 234)
    ... except TypeError as te:
    ...     print(te)
    algorithm name should be an instance of any in {None, str} but is int, \
namely '1'.
    >>> try:
    ...     MultiRunData("x x", "i", "f", "e", 234)
    ... except ValueError as ve:
    ...     print(ve)
    Invalid algorithm name 'x x'.
    >>> try:
    ...     MultiRunData("a", 5.5, "f", "e", 234)
    ... except TypeError as te:
    ...     print(te)
    instance name should be an instance of any in {None, str} but is float, \
namely '5.5'.
    >>> try:
    ...     MultiRunData("x", "a-i", "f", "e", 234)
    ... except ValueError as ve:
    ...     print(ve)
    Invalid instance name 'a-i'.
    >>> try:
    ...     MultiRunData("a", "i", True, "e", 234)
    ... except TypeError as te:
    ...     print(te)
    objective name should be an instance of any in {None, str} but is bool, \
namely 'True'.
    >>> try:
    ...     MultiRunData("xx", "i", "d'@f", "e", 234)
    ... except ValueError as ve:
    ...     print(ve)
    Invalid objective name "d'@f".
    >>> try:
    ...     MultiRunData("yy", "i", "f", -9.4, 234)
    ... except TypeError as te:
    ...     print(te)
    encoding name should be an instance of any in {None, str} but is float, \
namely '-9.4'.
    >>> try:
    ...     MultiRunData("xx", "i", "f", "e-{a", 234)
    ... except ValueError as ve:
    ...     print(ve)
    Invalid encoding name 'e-{a'.
    >>> try:
    ...     MultiRunData("x", "i", "f", "e", -1.234)
    ... except TypeError as te:
    ...     print(te)
    n should be an instance of int but is float, namely '-1.234'.
    >>> try:
    ...     MultiRunData("xx", "i", "f", "e", 1_000_000_000_000_000_000_000)
    ... except ValueError as ve:
    ...     print(ve)
    n=1000000000000000000000 is invalid, must be in 1..1000000000000000.
    """

    #: The algorithm that was applied, if the same over all runs.
    algorithm: str | None
    #: The problem instance that was solved, if the same over all runs.
    instance: str | None
    #: the name of the objective function, if the same over all runs
    objective: str | None
    #: the encoding, if any, or `None` if no encoding was used or if it was
    #: not the same over all runs
    encoding: str | None
    #: The number of runs over which the statistic information is computed.
    n: int

    def __init__(self, algorithm: str | None, instance: str | None,
                 objective: str | None, encoding: str | None, n: int):
        """
        Create the dataset of an experiment-setup combination.

        :param algorithm: the algorithm name, if all runs are with the same
            algorithm, `None` otherwise
        :param instance: the instance name, if all runs are on the same
            instance, `None` otherwise
        :param objective: the objective name, if all runs are on the same
            objective function, `None` otherwise
        :param encoding: the encoding name, if all runs are on the same
            encoding and an encoding was actually used, `None` otherwise
        :param n: the total number of runs
        """
        _set_name(self, algorithm, "algorithm", True, False)
        _set_name(self, instance, "instance", True, False)
        _set_name(self, objective, "objective", True, False)
        _set_name(self, encoding, "encoding", True, False)
        object.__setattr__(self, "n", check_int_range(
            n, "n", 1, 1_000_000_000_000_000))

    def _tuple(self) -> tuple[Any, ...]:
        """
        Get the tuple representation of this object used in comparisons.

        :return: the comparison-relevant data of this object in a tuple

        >>> MultiRunData("a", "i", "f", None, 3)._tuple()
        ('MultiRunData', 'a', 'i', 'f', '', 3, -1)
        >>> MultiRunData(None, "i", "f", "e", 31)._tuple()
        ('MultiRunData', '', 'i', 'f', 'e', 31, -1)
        >>> MultiRunData("x", None, "fy", "e1", 131)._tuple()
        ('MultiRunData', 'x', '', 'fy', 'e1', 131, -1)
        >>> MultiRunData("yx", "z", None, "xe1", 2131)._tuple()
        ('MultiRunData', 'yx', 'z', '', 'xe1', 2131, -1)
        """
        return (self.__class__.__name__,
                "" if self.algorithm is None else self.algorithm,
                "" if self.instance is None else self.instance,
                "" if self.objective is None else self.objective,
                "" if self.encoding is None else self.encoding,
                self.n, -1)


@dataclass(frozen=True, init=False, order=False, eq=False)
class MultiRun2DData(MultiRunData):
    """
    A multi-run data based on one time and one objective dimension.

    >>> p = MultiRun2DData("a", "i", "f", None, 3,
    ...                    TIME_UNIT_FES, F_NAME_SCALED)
    >>> p.instance
    'i'
    >>> p.algorithm
    'a'
    >>> p.objective
    'f'
    >>> print(p.encoding)
    None
    >>> p.n
    3
    >>> print(p.time_unit)
    FEs
    >>> print(p.f_name)
    scaledF
    >>> try:
    ...     MultiRun2DData("a", "i", "f", None, 3,
    ...                    3, F_NAME_SCALED)
    ... except TypeError as te:
    ...     print(te)
    time_unit should be an instance of str but is int, namely '3'.
    >>> try:
    ...     MultiRun2DData("a", "i", "f", None, 3,
    ...                    "sdfjsdf", F_NAME_SCALED)
    ... except ValueError as ve:
    ...     print(ve)
    Invalid time unit 'sdfjsdf', only 'FEs' and 'ms' are permitted.
    >>> try:
    ...     MultiRun2DData("a", "i", "f", None, 3,
    ...                    TIME_UNIT_FES, True)
    ... except TypeError as te:
    ...     print(te)
    f_name should be an instance of str but is bool, namely 'True'.
    >>> try:
    ...     MultiRun2DData("a", "i", "f", None, 3,
    ...                    TIME_UNIT_FES, "blablue")
    ... except ValueError as ve:
    ...     print(ve)
    Invalid f name 'blablue', only 'plainF', 'scaledF', and 'normalizedF' \
are permitted.
    """

    #: The unit of the time axis.
    time_unit: str
    #: the name of the objective value axis.
    f_name: str

    def __init__(self, algorithm: str | None, instance: str | None,
                 objective: str | None, encoding: str | None, n: int,
                 time_unit: str, f_name: str):
        """
        Create multi-run data based on one time and one objective dimension.

        :param algorithm: the algorithm name, if all runs are with the same
            algorithm
        :param instance: the instance name, if all runs are on the same
            instance
        :param objective: the objective name, if all runs are on the same
            objective function, `None` otherwise
        :param encoding: the encoding name, if all runs are on the same
            encoding and an encoding was actually used, `None` otherwise
        :param n: the total number of runs
        :param time_unit: the time unit
        :param f_name: the objective dimension name
        """
        super().__init__(algorithm, instance, objective, encoding, n)
        object.__setattr__(self, "time_unit", check_time_unit(time_unit))
        object.__setattr__(self, "f_name", check_f_name(f_name))

    def _tuple(self) -> tuple[Any, ...]:
        """
        Get the tuple representation of this object used in comparisons.

        :return: the comparison-relevant data of this object in a tuple

        >>> MultiRun2DData("a", "i", "f", None, 3,
        ...                TIME_UNIT_FES, F_NAME_SCALED)._tuple()
        ('MultiRun2DData', 'a', 'i', 'f', '', 3, -1, 'FEs', 'scaledF')
        >>> MultiRun2DData(None, "ix", None, "x", 43,
        ...                TIME_UNIT_MILLIS, F_NAME_RAW)._tuple()
        ('MultiRun2DData', '', 'ix', '', 'x', 43, -1, 'ms', 'plainF')
        >>> MultiRun2DData("xa", None, None, None, 143,
        ...                TIME_UNIT_MILLIS, F_NAME_NORMALIZED)._tuple()
        ('MultiRun2DData', 'xa', '', '', '', 143, -1, 'ms', 'normalizedF')
        """
        return (self.__class__.__name__,
                "" if self.algorithm is None else self.algorithm,
                "" if self.instance is None else self.instance,
                "" if self.objective is None else self.objective,
                "" if self.encoding is None else self.encoding,
                self.n, -1, self.time_unit, self.f_name)


def get_instance(obj: PerRunData | MultiRunData) -> str | None:
    """
    Get the instance of a given object.

    :param obj: the object
    :return: the instance string, or `None` if no instance is specified

    >>> p1 = MultiRunData("a", "i1", None, "x", 3)
    >>> get_instance(p1)
    'i1'
    >>> p2 = PerRunData("a", "i2", "f", "x", 31)
    >>> get_instance(p2)
    'i2'
    """
    return obj.instance


def get_algorithm(obj: PerRunData | MultiRunData) -> str | None:
    """
    Get the algorithm of a given object.

    :param obj: the object
    :return: the algorithm string, or `None` if no algorithm is specified

    >>> p1 = MultiRunData("a1", "i1", "f", "y", 3)
    >>> get_algorithm(p1)
    'a1'
    >>> p2 = PerRunData("a2", "i2", "y", None, 31)
    >>> get_algorithm(p2)
    'a2'
    """
    return obj.algorithm


def sort_key(obj: PerRunData | MultiRunData) -> tuple[Any, ...]:
    """
    Get the default sort key for the given object.

    The sort key is a tuple with well-defined field elements that should
    allow for a default and consistent sorting over many different elements of
    the experiment evaluation data API. Sorting should work also for lists
    containing elements of different classes.

    :param obj: the object
    :return: the sort key

    >>> p1 = MultiRunData("a1", "i1", "f", None, 3)
    >>> p2 = PerRunData("a2", "i2", "f", None, 31)
    >>> sort_key(p1) < sort_key(p2)
    True
    >>> sort_key(p1) >= sort_key(p2)
    False
    >>> p3 = MultiRun2DData("a", "i", "f", None, 3,
    ...                     TIME_UNIT_FES, F_NAME_SCALED)
    >>> sort_key(p3) < sort_key(p1)
    True
    >>> sort_key(p3) >= sort_key(p1)
    False
    """
    # noinspection PyProtectedMember
    return obj._tuple()


def motipy_footer_bottom_comments(
        _: Any, dest: Callable[[str], Any],
        additional: str | None = None) -> None:
    """
    Print the standard csv footer for moptipy.

    :param _: the setup object, ignored
    :param dest: the destination callable
    :param dest: the destination to write to
    :param additional: any additional output string

    >>> def __qpt(s: str):
    ...     print(s[:49])
    >>> motipy_footer_bottom_comments(None, __qpt, "bla")
    This data has been generated with moptipy version
    bla
    You can find moptipy at https://thomasweise.githu

    >>> motipy_footer_bottom_comments(None, __qpt, None)
    This data has been generated with moptipy version
    You can find moptipy at https://thomasweise.githu
    """
    dest("This data has been generated with moptipy version "
         f"{moptipy_version}.")
    if (additional is not None) and (str.__len__(additional) > 0):
        dest(additional)
    dest("You can find moptipy at https://thomasweise.github.io/mopitpy.")


#: a description of the algorithm field
DESC_ALGORITHM: Final[str] = "the name of the algorithm setup that was used."
#: a description of the instance field
DESC_INSTANCE: Final[str] = ("the name of the problem instance to which the "
                             "algorithm was applied.")
#: a description of the objective function field
DESC_OBJECTIVE_FUNCTION: Final[str] = \
    ("the name of the objective function (often also called fitness function "
     "or cost function) that was used to rate the solution quality.")
#: a description of the encoding field
DESC_ENCODING: Final[str] = \
    ("the name of the encoding, often also called genotype-phenotype mapping"
     ", used. In some problems, the search space on which the algorithm "
     "works is different from the space of possible solutions. For example, "
     "when solving a scheduling problem, maybe our optimization algorithm "
     "navigates in the space of permutations, but the solutions are Gantt "
     "charts. The encoding is the function that translates the points in "
     "the search space (e.g., permutations) to the points in the solution "
     "space (e.g., Gantt charts). Nothing if no encoding was used.")



"""Print a help screen."""


import argparse

from pycommons.io.arguments import make_argparser, make_epilog

from moptipy.version import __version__


def moptipy_argparser(file: str, description: str,
                      epilog: str) -> argparse.ArgumentParser:
    """
    Create an argument parser with default settings.

    :param file: the `__file__` special variable of the calling script
    :param description: the description string
    :param epilog: the epilogue string
    :returns: the argument parser

    >>> ap = moptipy_argparser(
    ...     __file__, "This is a test program.", "This is a test.")
    >>> isinstance(ap, argparse.ArgumentParser)
    True
    >>> "Copyright" in ap.epilog
    True
    """
    return make_argparser(
        file, description,
        make_epilog(epilog, 2022, 2024, "Thomas Weise",
                    url="https://thomasweise.github.io/moptipy",
                    email="tweise@hfuu.edu.cn, tweise@ustc.edu.cn"),
        __version__)



"""Utilities for interaction with numpy."""
from hashlib import sha512
from math import isfinite, isnan
from typing import Any, Final, cast

import numba  # type: ignore
import numpy as np
from numpy.random import PCG64, Generator, default_rng
from pycommons.io.csv import CSV_SEPARATOR
from pycommons.strings.string_conv import bool_to_str, num_to_str
from pycommons.types import check_int_range, type_error

#: All the numpy integer types and their ranges in increasing order of size.
#: The tuple contains alternating signed and unsigned types. It starts with
#: the smallest signed type, `numpy.int8` and ends with the largest unsigned
#: type `numpy.uint64`.
#: If we have a range `[min..max]` of valid value, then we can look up this
#: range and find the integer type with the smallest memory footprint to
#: accommodate it. This is what :func:`int_range_to_dtype` does.
__INTS_AND_RANGES: Final[tuple[tuple[np.dtype, int, int], ...]] = \
    tuple(sorted([
        (dtx, int(np.iinfo(dtx).min), int(np.iinfo(dtx).max))
        for dtx in cast(set[np.dtype], {np.dtype(bdt) for bdt in [
            int, np.int8, np.int16, np.uint8, np.uint16,
            np.int32, np.uint32, np.int64, np.uint64]})],
        key=lambda a: (a[2], a[1])))

#: The numpy integer data types.
INTS: Final[tuple[np.dtype, ...]] = tuple(a[0] for a in __INTS_AND_RANGES)

#: A map associating all numpy integer types associated to tuples
#: of their respective minimum and maximum value.
__NP_INTS_MAP: Final[dict[np.dtype, tuple[int, int]]] = \
    {a[0]: (a[1], a[2]) for a in __INTS_AND_RANGES}

#: The default integer type: the signed 64-bit integer.
DEFAULT_INT: Final[np.dtype] = INTS[-2]

#: The default unsigned integer type: an unsigned 64-bit integer.
DEFAULT_UNSIGNED_INT: Final[np.dtype] = INTS[-1]

#: The default boolean type.
DEFAULT_BOOL: Final[np.dtype] = np.dtype(np.bool_)

#: The default floating point type.
DEFAULT_FLOAT: Final[np.dtype] = np.dtype(float)

#: The default numerical types.
DEFAULT_NUMERICAL: Final[tuple[np.dtype, ...]] = (*list(INTS), DEFAULT_FLOAT)


def is_np_int(dtype: np.dtype) -> bool:
    """
    Check whether a :class:`numpy.dtype` is an integer type.

    :param dtype: the type

    >>> import numpy as npx
    >>> from moptipy.utils.nputils import is_np_int
    >>> print(is_np_int(npx.dtype(npx.int8)))
    True
    >>> print(is_np_int(npx.dtype(npx.uint16)))
    True
    >>> print(is_np_int(npx.dtype(npx.float64)))
    False
    """
    return dtype.kind in ("i", "u")


def is_np_float(dtype: np.dtype) -> bool:
    """
    Check whether a :class:`numpy.dtype` is a floating point type.

    :param dtype: the type

    >>> import numpy as npx
    >>> from moptipy.utils.nputils import is_np_float
    >>> print(is_np_float(npx.dtype(npx.int8)))
    False
    >>> print(is_np_float(npx.dtype(npx.uint16)))
    False
    >>> print(is_np_float(npx.dtype(npx.float64)))
    True
    """
    return dtype.kind == "f"


def int_range_to_dtype(min_value: int, max_value: int,
                       force_signed: bool = False) -> np.dtype:
    """
    Convert an integer range to an appropriate numpy data type.

    The returned type is as compact as possible and signed types are
    preferred over unsigned types.
    The returned :class:`numpy.dtype` will allow accommodating all values in
    the inclusive interval `min_value..max_value`.

    :param min_value: the minimum value
    :param max_value: the maximum value
    :param force_signed: enforce signed types
    :return: the numpy integer range
    :raises TypeError: if the parameters are not integers
    :raises ValueError: if the range is invalid, i.e., if `min_value` exceeds
        `max_value` or either of them exceeds the possible range of the
        largest numpy integers.

    >>> from moptipy.utils.nputils import int_range_to_dtype
    >>> print(int_range_to_dtype(0, 127))
    int8
    >>> print(int_range_to_dtype(0, 128))
    uint8
    >>> print(int_range_to_dtype(0, 128, True))
    int16
    >>> print(int_range_to_dtype(0, 32767))
    int16
    >>> print(int_range_to_dtype(0, 32768))
    uint16
    >>> print(int_range_to_dtype(0, 32768, True))
    int32
    >>> print(int_range_to_dtype(0, (2 ** 31) - 1))
    int32
    >>> print(int_range_to_dtype(0, 2 ** 31))
    uint32
    >>> print(int_range_to_dtype(0, 2 ** 31, True))
    int64
    >>> print(int_range_to_dtype(0, (2 ** 63) - 1))
    int64
    >>> print(int_range_to_dtype(0, 2 ** 63))
    uint64
    >>> print(int_range_to_dtype(0, (2 ** 64) - 1))
    uint64
    >>> try:
    ...     int_range_to_dtype(0, (2 ** 64) - 1, True)
    ... except ValueError as e:
    ...     print(e)
    Signed integer range cannot exceed -9223372036854775808..922337203685477\
5807, but 0..18446744073709551615 was specified.
    >>> try:
    ...     int_range_to_dtype(0, (2 ** 64) + 1)
    ... except ValueError as e:
    ...     print(e)
    max_value for unsigned integers must be <=18446744073709551615, but is \
18446744073709551617 for min_value=0.
    >>> try:
    ...     int_range_to_dtype(-1, (2 ** 64) - 1)
    ... except ValueError as e:
    ...     print(e)
    Signed integer range cannot exceed -9223372036854775808..922337203685477\
5807, but -1..18446744073709551615 was specified.
    >>> try:
    ...     int_range_to_dtype(-1.0, (2 ** 64) - 1)
    ... except TypeError as e:
    ...     print(e)
    min_value should be an instance of int but is float, namely '-1.0'.
    >>> try:
    ...     int_range_to_dtype(-1, 'a')
    ... except TypeError as e:
    ...     print(e)
    max_value should be an instance of int but is str, namely 'a'.
    """
    if not isinstance(min_value, int):
        raise type_error(min_value, "min_value", int)
    if not isinstance(max_value, int):
        raise type_error(max_value, "max_value", int)
    if min_value > max_value:
        raise ValueError(
            f"min_value must be <= max_value, but min_value={min_value} "
            f"and max_value={max_value} was provided.")

    use_min_value: Final[int] = -1 if force_signed and (min_value >= 0) \
        else min_value
    for t in __INTS_AND_RANGES:
        if (use_min_value >= t[1]) and (max_value <= t[2]):
            return t[0]

    if (min_value >= 0) and (not force_signed):
        raise ValueError(
            "max_value for unsigned integers must be <="
            f"{(__INTS_AND_RANGES[-1])[2]}, but is {max_value}"
            f" for min_value={min_value}.")

    raise ValueError(
        f"Signed integer range cannot exceed {__INTS_AND_RANGES[-2][1]}.."
        f"{__INTS_AND_RANGES[-2][2]}, but {min_value}..{max_value} "
        "was specified.")


def dtype_for_data(always_int: bool,
                   lower_bound: int | float,
                   upper_bound: int | float) -> np.dtype:
    """
    Obtain the most suitable numpy data type to represent the data.

    If the data is always integer, the smallest possible integer type will be
    sought using :func:`int_range_to_dtype`. If `always_int` is `True` and
    one or both of the bounds are infinite, then the largest available integer
    type is returned. If the bounds are finite but exceed the integer range,
    a `ValueError` is thrown. If the data is not always integer, the `float64`
    is returned.

    :param always_int: is the data always integer?
    :param lower_bound: the lower bound of the data, set to `-inf` if no lower
        bound is known and we should assume the full integer range
    :param upper_bound: the upper bound of the data, set to `-inf` if no upper
        bound is known and we should assume the full integer range
    :raises ValueError: if the `lower_bound > upper_bound` or any bound is
        `nan` or the integer bounds exceed the largest int range
    :raises TypeError: if, well, you provide parameters of the wrong types

    >>> print(dtype_for_data(True, 0, 127))
    int8
    >>> print(dtype_for_data(True, 0, 128))
    uint8
    >>> print(dtype_for_data(True, -1, 32767))
    int16
    >>> print(dtype_for_data(True, 0, 32768))
    uint16
    >>> print(dtype_for_data(True, -1, 32768))
    int32
    >>> print(dtype_for_data(True, 0, 65535))
    uint16
    >>> print(dtype_for_data(True, 0, 65536))
    int32
    >>> print(dtype_for_data(True, -1, 65535))
    int32
    >>> print(dtype_for_data(True, 0, (2 ** 31) - 1))
    int32
    >>> print(dtype_for_data(True, 0, 2 ** 31))
    uint32
    >>> print(dtype_for_data(True, -1, 2 ** 31))
    int64
    >>> print(dtype_for_data(True, 0, (2 ** 63) - 1))
    int64
    >>> print(dtype_for_data(True, 0, 2 ** 63))
    uint64
    >>> print(dtype_for_data(True, 0, (2 ** 63) + 1))
    uint64
    >>> print(dtype_for_data(True, 0, (2 ** 64) - 1))
    uint64
    >>> try:
    ...     dtype_for_data(True, 0, 2 ** 64)
    ... except ValueError as v:
    ...     print(v)
    max_value for unsigned integers must be <=18446744073709551615, but \
is 18446744073709551616 for min_value=0.
    >>> from math import inf, nan
    >>> print(dtype_for_data(True, 0, inf))
    uint64
    >>> print(dtype_for_data(True, -1, inf))
    int64
    >>> print(dtype_for_data(True, -inf, inf))
    int64
    >>> print(dtype_for_data(True, -inf, inf))
    int64
    >>> try:
    ...     dtype_for_data(True, 1, 0)
    ... except ValueError as v:
    ...     print(v)
    invalid bounds [1,0].
    >>> try:
    ...     dtype_for_data(False, 1, 0)
    ... except ValueError as v:
    ...     print(v)
    invalid bounds [1,0].
    >>> try:
    ...     dtype_for_data(True, 1, nan)
    ... except ValueError as v:
    ...     print(v)
    invalid bounds [1,nan].
    >>> try:
    ...     dtype_for_data(False, nan, 0)
    ... except ValueError as v:
    ...     print(v)
    invalid bounds [nan,0].
    >>> print(dtype_for_data(False, 1, 2))
    float64
    >>> try:
    ...     dtype_for_data(False, nan, '0')
    ... except TypeError as v:
    ...     print(v)
    upper_bound should be an instance of any in {float, int} but is str, \
namely '0'.
    >>> try:
    ...     dtype_for_data(True, 'x', 0)
    ... except TypeError as v:
    ...     print(v)
    lower_bound should be an instance of any in {float, int} but is str, \
namely 'x'.
    >>> try:
    ...     dtype_for_data(True, 1.0, 2.0)
    ... except TypeError as v:
    ...     print(v)
    finite lower_bound of always_int should be an instance of int but is \
float, namely '1.0'.
    >>> try:
    ...     dtype_for_data(True, 0, 2.0)
    ... except TypeError as v:
    ...     print(v)
    finite upper_bound of always_int should be an instance of int but is \
float, namely '2.0'.
    >>> try:
    ...     dtype_for_data(3, 0, 2)
    ... except TypeError as v:
    ...     print(v)
    always_int should be an instance of bool but is int, namely '3'.
    """
    if not isinstance(always_int, bool):
        raise type_error(always_int, "always_int", bool)
    if not isinstance(lower_bound, int | float):
        raise type_error(lower_bound, "lower_bound", (int, float))
    if not isinstance(upper_bound, int | float):
        raise type_error(upper_bound, "upper_bound", (int, float))
    if isnan(lower_bound) or isnan(upper_bound) or \
            (lower_bound > upper_bound):
        raise ValueError(f"invalid bounds [{lower_bound},{upper_bound}].")
    if always_int:
        if isfinite(lower_bound):
            if not isinstance(lower_bound, int):
                raise type_error(
                    lower_bound, "finite lower_bound of always_int", int)
            if isfinite(upper_bound):
                if not isinstance(upper_bound, int):
                    raise type_error(
                        upper_bound, "finite upper_bound of always_int", int)
                return int_range_to_dtype(int(lower_bound), int(upper_bound))
            if lower_bound >= 0:
                return DEFAULT_UNSIGNED_INT
        return DEFAULT_INT
    return DEFAULT_FLOAT


@numba.jit(forceobj=True)
def np_ints_max(shape, dtype: np.dtype = DEFAULT_INT) -> np.ndarray:
    """
    Create an integer array of the given length filled with the maximum value.

    :param shape: the requested shape
    :param dtype: the data type (defaults to 64bit integers)
    :return: the new array

    >>> import numpy as npx
    >>> from moptipy.utils.nputils import np_ints_max
    >>> print(np_ints_max(4, npx.dtype("uint8")))
    [255 255 255 255]
    """
    return np.full(shape=shape, fill_value=__NP_INTS_MAP[dtype][1],
                   dtype=dtype)


#: the default number of bytes for random seeds
__SEED_BYTES: Final[int] = 8
#: the minimum acceptable random seed
__MIN_RAND_SEED: Final[int] = 0
#: the maximum acceptable random seed
__MAX_RAND_SEED: Final[int] = int((1 << (__SEED_BYTES * 8)) - 1)


def rand_seed_generate(random: Generator = default_rng()) -> int:
    """
    Draw a (pseudo-random) random seed.

    This method either uses a provided random number generator `random` or a
    default generator. It draws 8 bytes from this generator and converts them
    to an unsigned (64 bit) integer big-endian style.

    :param random: the random number generator to be used to generate the seed
    :return: the random seed
    :raises TypeError: if `random` is specified but is not an instance of
        `Generator`

    >>> from numpy.random import default_rng as drg
    >>> rand_seed_generate(default_rng(100))
    10991970318022328789
    >>> rand_seed_generate(default_rng(100))
    10991970318022328789
    >>> rand_seed_generate(default_rng(10991970318022328789))
    11139051376468819756
    >>> rand_seed_generate(default_rng(10991970318022328789))
    11139051376468819756
    >>> rand_seed_generate(default_rng(11139051376468819756))
    16592984639586750386
    >>> rand_seed_generate(default_rng(11139051376468819756))
    16592984639586750386
    >>> rand_seed_generate(default_rng(16592984639586750386))
    12064014979695949294
    >>> rand_seed_generate(default_rng(16592984639586750386))
    12064014979695949294
    """
    if not isinstance(random, Generator):
        raise type_error(random, "random", Generator)
    return rand_seed_check(int.from_bytes(
        random.bytes(__SEED_BYTES), byteorder="big", signed=False))


def rand_seed_check(rand_seed: Any) -> int:
    """
    Make sure that a random seed is valid.

    :param rand_seed: the random seed to check
    :return: the rand seed

    :raises TypeError: if the random seed is not an `int`
    :raises ValueError: if the random seed is not valid

    >>> rand_seed_check(1)
    1
    >>> rand_seed_check(0)
    0
    >>> try:
    ...     rand_seed_check(-1)
    ... except ValueError as ve:
    ...     print(ve)
    rand_seed=-1 is invalid, must be in 0..18446744073709551615.
    >>> rand_seed_check(18446744073709551615)
    18446744073709551615
    >>> try:
    ...     rand_seed_check(18446744073709551616)
    ... except ValueError as ve:
    ...     print(ve)
    rand_seed=18446744073709551616 is invalid, must be in 0..\
18446744073709551615.
    >>> try:
    ...     rand_seed_check(1.2)
    ... except TypeError as te:
    ...     print(te)
    rand_seed should be an instance of int but is float, namely '1.2'.
    """
    return check_int_range(rand_seed, "rand_seed",
                           __MIN_RAND_SEED, __MAX_RAND_SEED)


def rand_generator(seed: int) -> Generator:
    """
    Instantiate a random number generator from a seed.

    :param seed: the random seed
    :return: the random number generator

    >>> type(rand_generator(1))
    <class 'numpy.random._generator.Generator'>
    >>> type(rand_generator(1).bit_generator)
    <class 'numpy.random._pcg64.PCG64'>
    >>> rand_generator(1).random() == rand_generator(1).random()
    True
    """
    return default_rng(rand_seed_check(seed))


def rand_seeds_from_str(string: str, n_seeds: int) -> list[int]:
    """
    Reproducibly generate `n_seeds` unique random seeds from a `string`.

    This function will produce a sorted sequence of `n_seeds` random seeds,
    each of which being an unsigned 64-bit integer, from the string passed in.
    The same string will always yield the same sequence reproducibly.
    Running the function twice with different values of `n_seeds` will result
    in the two sets of random seeds, where the larger one (for the larger
    value of `n_seeds`) contains all elements of the smaller one.

    This works as follows: First, we encode the string to an array of bytes
    using the UTF-8 encoding (`string.encode("utf8")`). Then, we compute the
    SHA-512 digest of this byte array (using `hashlib.sha512`).
    From this digest, we then use two chunks of 32 bytes (256 bit) to seed two
    :class:`~numpy.random.PCG64` random number generators. We then
    alternatingly draw seeds from these two generators using
    :func:`rand_seed_generate` until we have `n_seeds` unique values.

    This procedure is used in :func:`moptipy.api.experiment.run_experiment` to
    draw the random seeds for the algorithm runs to be performed. As `string`
    input, that method uses the string representation of the problem instance.
    This guarantees that all algorithms start with the same seeds on the same
    problems. It also guarantees that an experiment is repeatable, i.e., will
    use the same seeds when executed twice. Finally, it ensures that
    cherry-picking is impossible, as all seeds are fairly pseudo-random.

    1. Penny Pritzker and Willie E. May, editors, *Secure Hash Standard
       (SHS),* Federal Information Processing Standards Publication FIPS PUB
       180-4, Gaithersburg, MD, USA: National Institute of Standards and
       Technology, Information Technology Laboratory, August 2015.
       doi: https://dx.doi.org/10.6028/NIST.FIPS.180-4
       https://nvlpubs.nist.gov/nistpubs/FIPS/NIST.FIPS.180-4.pdf
    2. Unicode Consortium, editors, *The Unicode(R) Standard, Version
       15.0 - Core Specification,* Mountain View, CA, USA: Unicode, Inc.,
       September 2022, ISBN:978-1-936213-32-0,
       https://www.unicode.org/versions/Unicode15.0.0/
    3. NumPy Community, Permuted Congruential Generator (64-bit, PCG64), in
       *NumPy Reference, Release 1.23.0,* June 2022, Austin, TX, USA:
       NumFOCUS, Inc., https://numpy.org/doc/1.23/numpy-ref.pdf
    4. Melissa E. O'Neill: *PCG: A Family of Simple Fast Space-Efficient
       Statistically Good Algorithms for Random Number Generation,* Report
       HMC-CS-2014-0905, September 5, 2014, Claremont, CA, USA: Harvey Mudd
       College, Computer Science Department.
       https://www.cs.hmc.edu/tr/hmc-cs-2014-0905.pdf

    :param string: the string
    :param n_seeds: the number of seeds
    :return: a list of random seeds
    :raises TypeError: if the parameters do not follow the type contract
    :raises ValueError: if the parameter values are invalid

    >>> rand_seeds_from_str("hello world!", 1)
    [11688012229199056962]
    >>> rand_seeds_from_str("hello world!", 2)
    [3727742416375614079, 11688012229199056962]
    >>> rand_seeds_from_str("hello world!", 3)
    [3727742416375614079, 11688012229199056962, 17315292100125916507]

    >>> rand_seeds_from_str("metaheuristic optimization", 1)
    [12323230366215963648]
    >>> rand_seeds_from_str("metaheuristic optimization", 2)
    [12323230366215963648, 13673960948036381176]
    >>> rand_seeds_from_str("metaheuristic optimization", 3)
    [12323230366215963648, 13673960948036381176, 18426184104943646060]
    """
    if not isinstance(string, str):
        raise type_error(string, "string", str)
    if len(string) <= 0:
        raise ValueError("string must not be empty.")
    if not isinstance(n_seeds, int):
        raise type_error(n_seeds, "n_seeds", int)
    if n_seeds <= 0:
        raise ValueError(
            f"n_seeds must be positive, but is {n_seeds}.")

    seeds = bytearray(sha512(string.encode("utf8")).digest())
    seed1 = int.from_bytes(seeds[0:32], byteorder="big", signed=False)
    seed2 = int.from_bytes(seeds[32:64], byteorder="big", signed=False)
    del seeds

    # seed two PCG64 generators, each of which should use two 256 bit
    # numbers as seed
    g1 = Generator(PCG64(seed1))
    g2 = Generator(PCG64(seed2))

    generated: set[int] = set()
    while len(generated) < n_seeds:
        g1, g2 = g2, g1
        generated.add(rand_seed_generate(g1))

    result = list(generated)
    result.sort()

    if len(result) != n_seeds:
        raise ValueError("Failed to generate {n_seeds} unique seeds.")
    return result


@numba.njit(nogil=True)
def is_all_finite(a: np.ndarray) -> bool:
    """
    Check if an array is all finite.

    :param a: the input array
    :return: `True` if all elements in the array are finite, `False` otherwise

    >>> import numpy as npx
    >>> from moptipy.utils.nputils import is_all_finite
    >>> print(is_all_finite(npx.array([1.1, 2.1, 3])))
    True
    >>> print(is_all_finite(npx.array([1, 2, 3])))
    True
    >>> print(is_all_finite(npx.array([1.1, npx.inf, 3])))
    False
    """
    for x in a:  # noqa
        if not np.isfinite(x):  # noqa
            return False  # noqa
    return True  # noqa


#: the character identifying the numpy data type backing the space
KEY_NUMPY_TYPE: Final[str] = "dtype"


def numpy_type_to_str(dtype: np.dtype) -> str:
    """
    Convert a numpy data type to a string.

    :param dtype: the data type
    :returns: a string representation

    >>> import numpy as npx
    >>> numpy_type_to_str(npx.dtype(int))
    'l'
    >>> numpy_type_to_str(npx.dtype(float))
    'd'
    """
    return dtype.char


def np_to_py_number(number: Any) -> int | float:
    """
    Convert a scalar number from numpy to a corresponding Python type.

    :param number: the numpy number
    :returns: an integer or float representing the number

    >>> type(np_to_py_number(1))
    <class 'int'>
    >>> type(np_to_py_number(1.0))
    <class 'float'>
    >>> type(np_to_py_number(np.int8(1)))
    <class 'int'>
    >>> type(np_to_py_number(np.float64(1)))
    <class 'float'>
    >>> try:
    ...    np_to_py_number(np.complex64(1))
    ... except TypeError as te:
    ...    print(te)
    number should be an instance of any in {float, int, numpy.floating, \
numpy.integer} but is numpy.complex64, namely '(1+0j)'.
    """
    if isinstance(number, int):
        return number
    if isinstance(number, np.number):
        if isinstance(number, np.integer):
            return int(number)
        if isinstance(number, np.floating):
            return float(number)
    if isinstance(number, float):
        return number
    raise type_error(number, "number",
                     (int, float, np.integer, np.floating))


def array_to_str(data: np.ndarray) -> str:
    """
    Convert a numpy array to a string.

    This method represents a numpy array as a string.
    It makes sure to include all the information stored in the array and to
    represent it as compactly as possible.

    If the array has numerical values, it will use the default CSV separator.
    If the array contains Boolean values, it will use no separator at all.

    :param data: the data
    :returns: the string

    >>> import numpy as npx
    >>> array_to_str(npx.array([1, 2, 3]))
    '1;2;3'
    >>> array_to_str(npx.array([1, 2.2, 3]))
    '1;2.2;3'
    >>> array_to_str(npx.array([True, False, True]))
    'TFT'
    """
    if not isinstance(data, np.ndarray):
        raise type_error(data, "data", np.ndarray)
    k: Final[str] = data.dtype.kind
    if k in ("i", "u"):
        return CSV_SEPARATOR.join(map(str, data))
    if k == "f":
        return CSV_SEPARATOR.join(num_to_str(float(d)) for d in data)
    if k == "b":
        return "".join(bool_to_str(bool(d)) for d in data)
    raise ValueError(
        f"unsupported data kind {k!r} of type {str(data.dtype)!r}.")


@numba.njit(cache=True, inline="always")
def fill_in_canonical_permutation(a: np.ndarray) -> None:
    """
    Fill the canonical permutation into an array.

    >>> import numpy
    >>> arr = numpy.empty(10, int)
    >>> fill_in_canonical_permutation(arr)
    >>> print(arr)
    [0 1 2 3 4 5 6 7 8 9]
    """
    for i in range(len(a)):  # pylint: disable=C0200
        a[i] = i



"""
The modified Greedy (2+1) EAmod Evolutionary Algorithm.

The Greedy (2+1) EAmod maintains a population of two individuals. Both
solutions, `x0` and `x1`, are intiially sampled independently and at random.
Each iteration consists of two steps, a crossover step and a mutation step.
The binary operator (crossover) is only applied if `x0` and `x1` have the
same objective value to produce offspring `z1`. If `x0` and `x1` have
different objective values, then `z1` is set to be the better of the two
parents. Then, the final offspring `z2` is derived by applying the unary
mutation operator. If `z2` is at least as good as the better one of `x0`
and `x1`, then it will be accepted. If `x1` and `x0` are as same as
good, one of them is randomly chosen to be replaced by `z2`. Otherwise,
the worse one is replaced.

This is the implementation of a general black-box version of the
Greedy (2+1) GAmod by Carvalho Pinto and Doerr. The original algorithm is a
genetic algorithm, i.e., an EA with a bit string-based search space and a
mutation operator flipping a Binomially distributed number of bits and
performing uniform crossover. Here we implement is a general EA where you can
plug in any crossover or mutation operator. Furthermore, the algorithm by
Carvalho Pinto and Doerr is, in turn, a modified version of Sudhold's
Greedy (2+1) GA, with some improvements for efficiency.

1. Eduardo Carvalho Pinto and Carola Doerr. Towards a More Practice-Aware
   Runtime Analysis of Evolutionary Algorithms. 2017. arXiv:1812.00493v1
   [cs.NE] 3 Dec 2018. [Online]. http://arxiv.org/pdf/1812.00493.pdf.
2. Dirk Sudholt. Crossover Speeds up Building-Block Assembly. In Proceedings
   of the 14th Annual Conference on Genetic and Evolutionary Computation
   (GECCO'12), July 7-11, 2012, Philadelphia, Pennsylvania, USA,
   pages 689-702. ACM, 2012. https://doi.org/10.1145/2330163.2330260.
"""
from typing import Callable, Final

from numpy.random import Generator

from moptipy.api.algorithm import Algorithm2
from moptipy.api.operators import Op0, Op1, Op2
from moptipy.api.process import Process


class GreedyTwoPlusOneEAmod(Algorithm2):
    """The modified Greedy (2+1) Evolutionary Algorithm."""

    def solve(self, process: Process) -> None:
        """
        Apply the EA to an optimization problem.

        :param process: the black-box process object
        """
        # Omitted for brevity: store function references in variables
        random: Final[Generator] = process.get_random()  # random gen
        create: Final[Callable] = process.create  # create x container
        evaluate: Final[Callable] = process.evaluate  # the objective
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = self.op1.op1  # the unary operator
        op2: Final[Callable] = self.op2.op2  # the binary operator
        should_terminate: Final[Callable] = process.should_terminate
        equals: Final[Callable] = process.is_equal
        ri: Final[Callable] = random.integers

        x0 = create()  # allocate record for first solution
        op0(random, x0)  # randomly initialize first solution
        f0: int | float = evaluate(x0)  # evaluate first solution
        if should_terminate():  # should we quit?
            return  # yes.

        x1 = create()  # allocate record for first solution
        op0(random, x1)  # randomly initialize second solution
        f1: int | float = evaluate(x1)  # evaluate 2nd solution

        z1 = create()  # create record for result of binary operation
        z2 = create()  # create record for result of unary operation

        while not should_terminate():  # loop until budget is used up
            if f0 == f1:  # only perform binary operation if f0 == f1
                op2(random, z1, x0, x1)  # recombination
                p = z1  # input of unary operation comes from binary op
            else:
                if f0 > f1:  # swap x0 and x1 if x1 is better
                    f0, f1 = f1, f0  # swap objective values
                    x0, x1 = x1, x0  # swap solutions
                p = x0  # input of unary operation is best-so-far x
            op1(random, z2, p)  # apply unary operation
            if not (equals(z2, x0) or equals(z2, x1)):  # is z2 new?
                fnew = evaluate(z2)  # only then evaluate it
                if fnew <= f0:  # is it better or equal than x1
                    if (f1 > f0) or (ri(2) == 0):
                        z2, x1 = x1, z2  # swap z2 with x1 and
                        f1 = fnew  # and remember its quality
                    else:  # f1 == f2 and probability = 0.5
                        z2, x0 = x0, z2  # swap z2 with x0
                        f0 = fnew  # and remember its quality

    def __init__(self, op0: Op0, op1: Op1, op2: Op2) -> None:
        """
        Create the algorithm with nullary, unary, and binary search operator.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param op2: the binary search operator
        """
        super().__init__("greedy2plus1EAmod", op0, op1, op2)



"""Test all the example code in the project's examples directory."""


from pycommons.dev.tests.examples_in_dir import check_examples_in_dir
from pycommons.io.path import file_path


def test_examples_in_examples_directory() -> None:
    """Test all the examples in the examples directory."""
    # First, we resolve the directories
    check_examples_in_dir(file_path(__file__).up(2).resolve_inside(
        "examples"))



"""
Tournament selection with replacement in the tournaments.

For each slot in the destination, a tournament with
:attr:`~moptipy.algorithms.modules.selections.tournament_with_repl.\
TournamentWithReplacement.size`
randomly chosen participating solutions is conducted. The solution with
the best fitness wins and is copied to the destination. The solutions are
drawn with replacement for each tournament, meaning that one solution may
enter a given tournament several times. A solution may be selected multiple
times.

Tournament selection without replacement is implemented in
:mod:`moptipy.algorithms.modules.selections.tournament_without_repl`.

1. Peter J. B. Hancock. An Empirical Comparison of Selection Methods in
   Evolutionary Algorithms. In Terence Claus Fogarty, editor, *Selected Papers
   from the AISB Workshop on Evolutionary Computing (AISB EC'94),* April
   11-13, 1994, Leeds, UK, volume 865 of Lecture Notes in Computer Science,
   pages 80-94, Berlin/Heidelberg, Germany: Springer, ISBN: 978-3-540-58483-4.
   https://dx.doi.org/10.1007/3-540-58483-8_7. Conference organized by the
   Society for the Study of Artificial Intelligence and Simulation of
   Behaviour (AISB).
2. Uday Kumar Chakraborty and Kalyanmoy Deb and Mandira Chakraborty. Analysis
   of Selection Algorithms: A Markov Chain Approach. *Evolutionary
   Computation,* 4(2):133-167. Summer 1996. Cambridge, MA, USA: MIT Press.
   doi:10.1162/evco.1996.4.2.133.
   https://dl.acm.org/doi/pdf/10.1162/evco.1996.4.2.133
3. Tobias Blickle and Lothar Thiele. A Comparison of Selection Schemes used in
   Genetic Algorithms. Second edition, December 1995. TIK-Report 11 from the
   Eidgenössische Technische Hochschule (ETH) Zürich, Department of Electrical
   Engineering, Computer Engineering and Networks Laboratory (TIK), Zürich,
   Switzerland. ftp://ftp.tik.ee.ethz.ch/pub/publications/TIK-Report11.ps
4. Kumara Sastry and David Edward Goldberg. Modeling Tournament Selection with
   Replacement using Apparent Added Noise. In Lee Spector, Erik D. Goodman,
   Annie Wu, William Benjamin Langdon, and Hans-Michael Voigt, eds.,
   *Proceedings of the 3rd Annual Conference on Genetic and Evolutionary
   Computation (GECCO'01)*, July 7-11, 2001, San Francisco, CA, USA, page 781.
   San Francisco, CA, United States: Morgan Kaufmann Publishers Inc.
   ISBN: 978-1-55860-774-3. https://dl.acm.org/doi/pdf/10.5555/2955239.2955378
"""

from math import inf
from typing import Any, Callable, Final

from numpy.random import Generator
from pycommons.types import check_int_range

from moptipy.algorithms.modules.selection import FitnessRecord, Selection
from moptipy.utils.logger import KeyValueLogSection


# start book
class TournamentWithReplacement(Selection):
    """Tournament selection with replacement in the tournaments."""

# end book
    def __init__(self, size: int = 2) -> None:
        """
        Create the tournament selection with replacement method.

        :param size: the size of the tournaments
        """
        super().__init__()
        #: the tournament size
        self.size: Final[int] = check_int_range(size, "tournament size", 1)

# start book
    def select(self, source: list[FitnessRecord],
               dest: Callable[[FitnessRecord], Any],
               n: int, random: Generator) -> None:
        """
        Perform tournament with replacement.

        :param source: the list with the records to select from
        :param dest: the destination collector to invoke for each selected
            record
        :param n: the number of records to select
        :param random: the random number generator
        """
        size: Final[int] = self.size  # the tournament size
        m: Final[int] = len(source)  # number of elements to select from
        ri: Final[Callable] = random.integers  # fast call to random.integers
        for _ in range(n):  # conduct n tournaments
            best: FitnessRecord | None = None  # best competitor
            best_fitness: int | float = inf  # best fitness, initial infinite
            for __ in range(size):  # perform tournament
                rec = source[ri(m)]  # get contestant record from source
                rec_fitness = rec.fitness  # get its fitness
                if rec_fitness <= best_fitness:  # if better or equal...
                    best = rec  # ... rec becomes the new best record
                    best_fitness = rec_fitness  # and remember fitness
            dest(best)  # at end of the tournament, send best to dest
        # end book

    def __str__(self):
        """
        Get the name of the tournament selection algorithm.

        :return: the name of the tournament selection algorithm
        """
        return f"tour{self.size}r"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("size", self.size)



"""
A search space for signed permutations of a base string.

The base string can be a string where each element occurs exactly once,
e.g., `[1, 2, 3]` or `[7, 2, 5]`. It could also be a string where some or
all elements occur multiple times, e.g., a permutation with repetition, such
as `[1, 7, 7, 4, 5, 3, 3, 3]` or `[3, 3, 1, 1, 2, 2]`. Each element may then
occur either in its original (positive) value in a string or in its negated
(negative) form.

Notice that a signed permutation can never include the element `0`. Therefore,
some methods have different semantics compared to those in class
:class:`~moptipy.spaces.permutations.Permutations`.

Search operators working on elements of this space are given in module
:mod:`moptipy.operators.signed_permutations`, but several of those in
:mod:`moptipy.operators.permutations` should work as well.
"""
from math import factorial
from typing import Counter, Final, Iterable

import numpy as np
from pycommons.types import check_int_range, type_error

from moptipy.spaces.intspace import IntSpace
from moptipy.spaces.permutations import KEY_BASE_STRING, KEY_REPETITIONS
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import array_to_str
from moptipy.utils.strings import sanitize_name

#: the unsigned minimum
KEY_UNSIGNED_MIN: Final[str] = "unsignedMin"


class SignedPermutations(IntSpace):  # +book
    """
    Signed permutations of a base string stored as :class:`numpy.ndarray`.

    This class includes standard permutations of the form 1, 2, ..., n,
    but also permutations with repetitions.
    The idea is that a base string is defined, i.e., an array of integer
    values. In this array, some values may appear twice, some may be missing.
    For example, `[1, 3, 5, 5, 7]` is a proper base string. The space defined
    over this base string would then contain values such as `[1, 3, 5, 5, 7]`,
    `[1, 5, 3, 5, 7]`, `[7, 5, 5, 3, 1]` and so on. Basically, it will contain
    all strings that can be created by shuffling the base string *and/or*
    negating some or all of the values, i.e., `[-1, 3, -5, 5, -7]` would also
    be possible. These strings have in common that they contain exactly all the
    values from the base string and contain them exactly as same as often as
    they appear in the base string. They can never contain `0`.

    The space defined upon the above base string contains
    (2 ** 5) * 5! / (1! * 1! * 2! * 1!) = 32 * 120 / 2 = 1920
    different strings.

    The permutation space defined above can be created as follows:

    >>> perm = SignedPermutations([1, 3, 5, 5, 7])
    >>> print(perm.to_str(perm.blueprint))
    1;3;5;5;7
    >>> print(perm)
    signedPermOfString
    >>> print(perm.n_points())
    1920

    Another example is this:

    >>> perm = SignedPermutations((1, 2, 3, 3, 2))
    >>> print(perm.to_str(perm.blueprint))
    1;2;2;3;3
    >>> print(perm)
    signedPermOfString
    >>> print(perm.n_points())
    960

    If you want to create a permutation with repetitions, e.g.,
    where each of the n=4 values from 1 to 4 appear exactly 3 times,
    you can use the utility method `with_repetitions`:

    >>> perm = SignedPermutations.with_repetitions(4, 3)
    >>> print(perm.to_str(perm.blueprint))
    1;1;1;2;2;2;3;3;3;4;4;4
    >>> print(perm)
    signedPerm4w3r
    >>> print(perm.n_points())
    1513881600

    If you instead want to create standard permutations, i.e., where
    each value from 1 to n appears exactly once, you would do:

    >>> perm = SignedPermutations.standard(5)
    >>> print(perm.to_str(perm.blueprint))
    1;2;3;4;5
    >>> print(perm)
    signedPerm5
    >>> print(perm.n_points())
    3840

    Different from normal permutations with repetitions, it is allowed
    that signed permutations with repetitions contain the same element,
    as long as their length is larger than 1.

    >>> perm = SignedPermutations([1, 1])
    >>> print(perm.to_str(perm.blueprint))
    1;1
    >>> x = perm.create()
    >>> perm.validate(x)
    >>> x[0] = -1
    >>> perm.validate(x)
    >>> x[1] = -1
    >>> perm.validate(x)
    >>> x[0] = 1
    >>> perm.validate(x)

    >>> try:
    ...     perm = SignedPermutations([1])
    ... except ValueError as ve:
    ...     print(ve)
    base_string must contain at least two different elements or have at \
least length 2, but is [1].
    """

    def __init__(self, base_string: Iterable[int]) -> None:  # +book
        """
        Create the space of signed permutations of a base string.

        :param base_string: an iterable of integer to denote the base string
        :raises TypeError: if one of the parameters has the wrong type
        :raises ValueError: if the parameters have the wrong value
        """
        if not isinstance(base_string, Iterable):
            raise type_error(base_string, "base_string", Iterable)

        string: Final[list[int]] = sorted(abs(i) for i in base_string)
        total: Final[int] = len(string)
        if total <= 0:
            raise ValueError(
                f"base string must not be empty, but is {base_string}.")

        # get data ranges
        self.__shape: Final[Counter[int]] = Counter[int]()
        minimum: int = string[0]
        maximum: int = string[0]
        for i in string:
            if not isinstance(i, int):
                raise type_error(i, "element of base_string", int)
            self.__shape[i] += 1
            minimum = min(minimum, i)
            maximum = max(maximum, i)

        # checking that the data is not empty
        different: Final[int] = len(self.__shape)
        if (total <= 1) and (different <= 1):
            raise ValueError("base_string must contain at least two different"
                             " elements or have at least length 2, but is "
                             f"{base_string}.")
        if minimum <= 0:
            raise ValueError(f"base_string must only contain positive values,"
                             f" but minimum is {minimum}.")

        # creating super class
        super().__init__(total, -maximum, maximum)

        #: a numpy array of the right type with the base string
        self.blueprint: Final[np.ndarray] = \
            np.array(string, dtype=self.dtype)

        npoints: int = factorial(total)
        rep: int | None = self.__shape.get(minimum)
        for v in self.__shape.values():
            npoints = npoints // factorial(v)
            if v != rep:
                rep = None
        #: the exact number of different signed permutations
        self.__n_points = (2 ** len(string)) * npoints

        #: the number of repetitions if all elements occur as same
        #: as often, or None otherwise
        self.__repetitions: Final[int | None] = rep

        #: the unsigned minimum
        self.unsigned_min_value: Final[int] = minimum

    def has_repetitions(self) -> bool:
        """
        Return whether elements occur repeatedly in the base string.

        :returns: `True` if at least one element occurs more than once,
            `False` otherwise
        """
        return (self.__repetitions is None) or (self.__repetitions > 1)

    def n(self) -> int:
        """
        Get the number of different values in the base string.

        :returns: the number of different values in the base string
        """
        return len(self.__shape)

    def is_dense(self) -> bool:
        """
        Check if all values in `min..max` appear in the permutation.

        :returns: `True` if the permutation is dense in the sense that
            all values from `self.min_value` to `self.max_value` appear.
        """
        return len(self.__shape) == (
            self.max_value - self.unsigned_min_value + 1)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this space to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)

        logger.key_value(KEY_UNSIGNED_MIN, self.unsigned_min_value)

        reps: Final[int | None] = self.__repetitions
        if reps:
            logger.key_value(KEY_REPETITIONS, reps)
            if self.is_dense():
                return
        logger.key_value(KEY_BASE_STRING,
                         array_to_str(self.blueprint))

    def create(self) -> np.ndarray:  # +book
        r"""
        Create a permutation equal to the base string.

        The result is of the form [1, 1, 2, 2, 2, 3, 3...].

        :return: the permutation of the base string

        >>> perm = SignedPermutations([1, 5, 2, 2, 4, 3, 4])
        >>> x = perm.create()
        >>> print(perm.to_str(x))
        1;2;2;3;4;4;5
        """
        return self.blueprint.copy()  # Create copy of the blueprint.

    def validate(self, x: np.ndarray) -> None:
        """
        Validate a signed permutation of the base string.

        :param x: the integer string
        :raises TypeError: if the string is not an :class:`numpy.ndarray`.
        :raises ValueError: if the element is not a valid element of this
            permutation space, e.g., if the shape or data type of `x` is
            wrong, if an element is outside of the bounds, or if an
            element does not occur exactly the prescribed amount of times
        """
        super().validate(x)
        counts: Counter[int] = Counter[int]()
        for xx in x:
            if xx == 0:
                raise ValueError("0 is not permitted in signed permutations.")
            counts[abs(xx)] += 1

        if counts != self.__shape:
            for key in sorted(set(counts.keys()).union(
                    set(self.__shape.keys()))):
                exp = self.__shape.get(key, 0)
                found = counts.get(key, 0)
                if found != exp:
                    raise ValueError(
                        f"Expected to find {key} exactly {exp} times "
                        f"but found it {found} times.")

    def n_points(self) -> int:
        """
        Get the number of possible different permutations of the base string.

        :return: (2 ** dimension) * factorial(dimension) /
            Product(factorial(count(e)) for all e)

        >>> print(SignedPermutations([1, 2, 3, 1, 2, 3]).n_points())
        5760
        """
        return self.__n_points

    def __str__(self) -> str:
        """
        Get the name of this space.

        :return: "perm" + blueprint string

        >>> print(SignedPermutations([3, 1, 5, 2, 1]))
        signedPermOfString
        >>> print(SignedPermutations([3, 2, 3, 1, 1, 2]))
        signedPerm3w2r
        >>> print(SignedPermutations([4, 2, 1, 3]))
        signedPerm4
        >>> print(SignedPermutations([3, 2, 4, 2, 3, 2, 4, 3, 4]))
        signedPerm2to4w3r
        """
        minimum: Final[int] = self.unsigned_min_value
        maximum: Final[int] = self.max_value
        reps: Final[int | None] = self.__repetitions
        different: Final[int] = self.n()
        if reps and (different != (self.dimension // reps)):
            raise ValueError(f"huh? {different} != {self.dimension} / {reps}")
        all_values: Final[bool] = self.is_dense()
        min_is_1: Final[bool] = minimum == 1

        if reps:
            if reps == 1:
                if all_values:
                    if min_is_1:
                        return f"signedPerm{different}"
                    return sanitize_name(f"signedPerm{minimum}to{maximum}")
            elif all_values:  # repetitions != 1
                if min_is_1:
                    return f"signedPerm{different}w{reps}r"
                return sanitize_name(f"signedPerm{minimum}to{maximum}w{reps}r")
        return "signedPermOfString"

    @staticmethod
    def standard(n: int) -> "SignedPermutations":
        """
        Create a space for permutations of `1..n`.

        Notice that this method is different compared to the method
        :meth:`~moptipy.spaces.permutations.Permutations.standard` of class
        :class:`~moptipy.spaces.permutations.Permutations` in that it starts
        the permutations at `1`, not at `0`.

        :param n: the range of the values
        :returns: the permutations space
        """
        return SignedPermutations(range(
            1, 1 + check_int_range(n, "n", 2, 100_000_000)))

    @staticmethod  # +book
    def with_repetitions(n: int, repetitions: int) -> "SignedPermutations":
        """
        Create a space for permutations of `1..n` with repetitions.

        Notice that this method is different compared to the method
        :meth:`~moptipy.spaces.permutations.Permutations.with_repetitions` of
        class :class:`~moptipy.spaces.permutations.Permutations` in that it
        starts the permutations at `1`, not at `0`.

        :param n: the range of the values
        :param repetitions: how often each value occurs
        :returns: the permutations space
        """
        check_int_range(repetitions, "repetitions", 1, 1_000_000)
        if repetitions <= 1:
            return SignedPermutations.standard(n)
        check_int_range(n, "n", 1, 100_000_000)
        return SignedPermutations(list(range(1, 1 + n)) * repetitions)
